#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import reduce

import numpy as np

import piquasso as pq

import pytest

from scipy.linalg import block_diag

from piquasso._math.transformations import xxpp_to_xpxp_indices, xpxp_to_xxpp_indices

from piquasso._math.linalg import is_selfadjoint, is_skew_symmetric
from piquasso._math.validations import all_in_interval

from piquasso.fermionic._utils import get_omega, binary_to_fock_indices


for_all_connectors = pytest.mark.parametrize(
    "connector", [pq.NumpyConnector(), pq.JaxConnector()]
)


@for_all_connectors
def test_vacuum_covariance_matrix(connector):
    d = 3

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    state.validate()

    assert type(state) is pq.fermionic.GaussianState

    J = np.array([[0, 1], [-1, 0]])

    assert np.allclose(state.covariance_matrix, block_diag(*[J] * d))


@for_all_connectors
def test_vacuum_correlation_matrix(connector):
    d = 3

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    ident = np.identity(d)
    zeros = np.zeros_like(ident)

    assert np.allclose(
        state.correlation_matrix, np.block([[zeros, zeros], [zeros, ident]])
    )


@pytest.mark.monkey
@for_all_connectors
def test_vacuum_evolved_with_GaussianHamiltonian_is_valid(
    connector,
    generate_fermionic_gaussian_hamiltonian,
):
    d = 2

    hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    state.validate()


@pytest.mark.monkey
@for_all_connectors
def test_vacuum_evolved_with_GaussianHamiltonian_correlation_matrix_random(
    connector,
    generate_fermionic_gaussian_hamiltonian,
):
    d = 3

    hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    gamma = state.correlation_matrix

    gamma11 = gamma[:d, :d]
    gamma12 = gamma[:d, d:]
    gamma21 = gamma[d:, :d]
    gamma22 = gamma[d:, d:]

    assert is_selfadjoint(gamma11)
    assert is_selfadjoint(gamma22)

    assert is_skew_symmetric(gamma12)
    assert is_skew_symmetric(gamma21)

    assert np.allclose(gamma11, np.identity(d) - gamma22.conj())
    assert np.allclose(gamma12, -gamma21.conj())

    assert np.allclose(
        gamma @ gamma, gamma
    ), "The correlation matrix is pure, must be projector"
    assert np.isclose(
        np.trace(gamma), d
    ), "The correlation matrix must be a projector of rank 3"


@pytest.mark.monkey
@for_all_connectors
def test_GaussianHamiltonian_covariance_and_correlation_matrix_equivalence(
    connector,
    generate_fermionic_gaussian_hamiltonian,
):
    d = 3

    hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    omega = get_omega(d)

    indices = xpxp_to_xxpp_indices(d)

    assert np.allclose(
        state.covariance_matrix[np.ix_(indices, indices)],
        -1j
        * omega
        @ (2 * state.correlation_matrix - np.identity(2 * d))
        @ omega.conj().T,
    ), "Eq. (45) from https://arxiv.org/pdf/2111.08343 should hold"


@for_all_connectors
def test_vacuum_correlation_matrix_density_matrix_equivalence(
    connector,
    get_ladder_operators,
):
    d = 3

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    fs = get_ladder_operators(d)

    rho = state.density_matrix

    correlation_matrix = state.correlation_matrix

    expected_correlation_matrix = np.empty_like(correlation_matrix)

    for i in range(2 * d):
        for j in range(2 * d):
            expected_correlation_matrix[i, j] = np.trace(rho @ fs[i].T @ fs[j])

    assert np.allclose(correlation_matrix, expected_correlation_matrix)


@pytest.mark.monkey
@for_all_connectors
def test_correlation_matrix_density_matrix_equivalence_random(
    connector,
    generate_fermionic_gaussian_hamiltonian,
    get_ladder_operators,
):
    d = 3

    hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    fs = get_ladder_operators(d)

    rho = state.density_matrix

    correlation_matrix = state.correlation_matrix

    expected_correlation_matrix = np.empty_like(correlation_matrix)

    for i in range(2 * d):
        for j in range(2 * d):
            expected_correlation_matrix[i, j] = np.trace(rho @ fs[i].T @ fs[j])

    assert np.allclose(correlation_matrix, expected_correlation_matrix)


@for_all_connectors
def test_vacuum_covariance_matrix_with_majorana_operators_and_density_matrix(
    connector, get_majorana_operators
):
    d = 3

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    m = get_majorana_operators(d)

    covariance_matrix = state.covariance_matrix

    rho = state.density_matrix

    for i in range(2 * d):
        for j in range(2 * d):
            assert np.isclose(
                covariance_matrix[i, j],
                -1j * np.trace(rho @ (m[i] @ m[j] - m[j] @ m[i])) / 2,
            )


@pytest.mark.monkey
@for_all_connectors
def test_covariance_matrix_with_majorana_operators_and_density_matrix(
    connector, generate_fermionic_gaussian_hamiltonian, get_majorana_operators
):
    d = 3

    hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    m = get_majorana_operators(d)

    covariance_matrix = state.covariance_matrix

    rho = state.density_matrix

    expected_covariance_matrix = np.empty(shape=(2 * d, 2 * d), dtype=rho.dtype)

    for i in range(2 * d):
        for j in range(2 * d):
            expected_covariance_matrix[i, j] = (
                -1j * np.trace(rho @ (m[i] @ m[j] - m[j] @ m[i])) / 2
            )

    assert np.allclose(covariance_matrix, expected_covariance_matrix)


@pytest.mark.monkey
@for_all_connectors
def test_GaussianHamiltonian_correlation_matrix_and_maj_equivalence(
    connector,
    generate_fermionic_gaussian_hamiltonian,
):
    d = 3

    hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    omega = get_omega(d)

    assert np.allclose(
        state.maj_correlation_matrix,
        omega @ state.correlation_matrix @ omega.conj().T,
    ), "Eq. (43) from https://arxiv.org/pdf/2111.08343 should hold"


@for_all_connectors
def test_vacuum_evolved_with_GaussianHamiltonian_correlation_matrix(connector):
    d = 3

    A = np.array([[1, 2j, 3j], [-2j, 5, 6], [-3j, 6, 7]])
    B = np.array([[0, 1j, 2], [-1j, 0, 3], [-2, -3, 0]])

    hamiltonian = np.block([[-A.conj(), B], [-B.conj(), A]])

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state
    state.validate()

    assert np.allclose(
        state.correlation_matrix,
        np.array(
            [
                [
                    0.33416157 - 0.0j,
                    -0.0401874 + 0.01146957j,
                    -0.05663529 - 0.00376422j,
                    0.0 + 0.0j,
                    -0.19852528 + 0.31881949j,
                    0.05833792 - 0.27031065j,
                ],
                [
                    -0.0401874 - 0.01146957j,
                    0.23155699 - 0.0j,
                    -0.15017863 - 0.05386455j,
                    0.19852528 - 0.31881949j,
                    0.0 + 0.0j,
                    -0.04634951 + 0.08677748j,
                ],
                [
                    -0.05663529 + 0.00376422j,
                    -0.15017863 + 0.05386455j,
                    0.13234052 - 0.0j,
                    -0.05833792 + 0.27031065j,
                    0.04634951 - 0.08677748j,
                    0.0 - 0.0j,
                ],
                [
                    -0.0 + 0.0j,
                    0.19852528 + 0.31881949j,
                    -0.05833792 - 0.27031065j,
                    0.66583843 - 0.0j,
                    0.0401874 + 0.01146957j,
                    0.05663529 - 0.00376422j,
                ],
                [
                    -0.19852528 - 0.31881949j,
                    -0.0 + 0.0j,
                    0.04634951 + 0.08677748j,
                    0.0401874 - 0.01146957j,
                    0.76844301 - 0.0j,
                    0.15017863 - 0.05386455j,
                ],
                [
                    0.05833792 + 0.27031065j,
                    -0.04634951 - 0.08677748j,
                    -0.0 - 0.0j,
                    0.05663529 + 0.00376422j,
                    0.15017863 + 0.05386455j,
                    0.86765948 - 0.0j,
                ],
            ]
        ),
    )


@for_all_connectors
def test_get_majorana_monomial_expectation_value_2_terms_consecutive(
    connector, get_majorana_operators
):
    d = 3

    A = np.array([[1, 2j, 3j], [-2j, 5, 6], [-3j, 6, 7]])
    B = np.array([[0, 1j, 2], [-1j, 0, 3], [-2, -3, 0]])

    hamiltonian = np.block([[-A.conj(), B], [-B.conj(), A]])

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state
    state.validate()

    density_matrix = state.density_matrix

    m = get_majorana_operators(d)

    assert np.isclose(
        state.get_majorana_monomial_expectation_value([0, 1]),
        np.trace(density_matrix @ m[0] @ m[1]),
    )


@for_all_connectors
def test_get_majorana_monomial_expectation_value_2_terms(
    connector, get_majorana_operators
):
    d = 3

    A = np.array([[1, 2j, 3j], [-2j, 5, 6], [-3j, 6, 7]])
    B = np.array([[0, 1j, 2], [-1j, 0, 3], [-2, -3, 0]])

    hamiltonian = np.block([[-A.conj(), B], [-B.conj(), A]])

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state
    state.validate()

    density_matrix = state.density_matrix

    m = get_majorana_operators(d)

    assert np.isclose(
        state.get_majorana_monomial_expectation_value([1, 3]),
        np.trace(density_matrix @ m[1] @ m[3]),
    )


@for_all_connectors
def test_get_majorana_monomial_expectation_value_4_terms(
    connector, get_majorana_operators
):
    d = 3

    A = np.array([[1, 2j, 3j], [-2j, 5, 6], [-3j, 6, 7]])
    B = np.array([[0, 1j, 2], [-1j, 0, 3], [-2, -3, 0]])

    hamiltonian = np.block([[-A.conj(), B], [-B.conj(), A]])

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state
    state.validate()

    density_matrix = state.density_matrix

    m = get_majorana_operators(d)

    assert np.isclose(
        state.get_majorana_monomial_expectation_value([0, 1, 3, 4]),
        np.trace(density_matrix @ m[0] @ m[1] @ m[3] @ m[4]),
    )


@for_all_connectors
def test_get_majorana_monomial_expectation_value_4_terms_consecutive(
    connector, get_majorana_operators
):
    d = 3

    A = np.array([[1, 2j, 3j], [-2j, 5, 6], [-3j, 6, 7]])
    B = np.array([[0, 1j, 2], [-1j, 0, 3], [-2, -3, 0]])

    hamiltonian = np.block([[-A.conj(), B], [-B.conj(), A]])

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state
    state.validate()

    density_matrix = state.density_matrix

    m = get_majorana_operators(d)

    assert np.isclose(
        state.get_majorana_monomial_expectation_value([0, 1, 2, 3]),
        np.trace(density_matrix @ m[0] @ m[1] @ m[2] @ m[3]),
    )


@for_all_connectors
def test_get_majorana_monomial_expectation_value_4_terms_flipped(
    connector, get_majorana_operators
):
    d = 3

    A = np.array([[1, 2j, 3j], [-2j, 5, 6], [-3j, 6, 7]])
    B = np.array([[0, 1j, 2], [-1j, 0, 3], [-2, -3, 0]])

    hamiltonian = np.block([[-A.conj(), B], [-B.conj(), A]])

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state
    state.validate()

    density_matrix = state.density_matrix

    m = get_majorana_operators(d)

    assert np.isclose(
        state.get_majorana_monomial_expectation_value([0, 2, 1, 3]),
        np.trace(density_matrix @ m[0] @ m[2] @ m[1] @ m[3]),
    )

    assert np.isclose(
        state.get_majorana_monomial_expectation_value([0, 1, 2, 3]),
        -state.get_majorana_monomial_expectation_value([0, 2, 1, 3]),
    ), "Flips should change the sign."


@for_all_connectors
def test_get_majorana_monomial_expectation_value_with_duplicates(
    connector, get_majorana_operators
):
    d = 3

    A = np.array([[1, 2j, 3j], [-2j, 5, 6], [-3j, 6, 7]])
    B = np.array([[0, 1j, 2], [-1j, 0, 3], [-2, -3, 0]])

    hamiltonian = np.block([[-A.conj(), B], [-B.conj(), A]])

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state
    state.validate()

    density_matrix = state.density_matrix

    m = get_majorana_operators(d)

    assert np.isclose(
        state.get_majorana_monomial_expectation_value([0, 2, 0, 1, 2, 3]),
        np.trace(density_matrix @ m[0] @ m[2] @ m[0] @ m[1] @ m[2] @ m[3]),
    )


@for_all_connectors
def test_get_majorana_monomial_expectation_value_with_triple(
    connector, get_majorana_operators
):
    d = 3

    A = np.array([[1, 2j, 3j], [-2j, 5, 6], [-3j, 6, 7]])
    B = np.array([[0, 1j, 2], [-1j, 0, 3], [-2, -3, 0]])

    hamiltonian = np.block([[-A.conj(), B], [-B.conj(), A]])

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state
    state.validate()

    density_matrix = state.density_matrix

    m = get_majorana_operators(d)

    assert np.isclose(
        state.get_majorana_monomial_expectation_value([5, 5, 0, 3, 0, 0, 1, 4]),
        np.trace(
            density_matrix @ m[5] @ m[5] @ m[0] @ m[3] @ m[0] @ m[0] @ m[1] @ m[4]
        ),
    )


@pytest.mark.monkey
@for_all_connectors
def test_get_majorana_monomial_expectation_value_random_without_multiplicities(
    connector, get_majorana_operators, generate_fermionic_gaussian_hamiltonian
):
    d = 3

    parent_hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    program = pq.Program(
        [pq.fermionic.ParentHamiltonian(hamiltonian=parent_hamiltonian)]
    )
    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)
    state = simulator.execute(program).state
    initial_length = np.random.randint(1, 10)

    indices = np.random.randint(0, 2 * d, initial_length)

    indices = list(set(indices))  # Remove duplicates

    density_matrix = state.density_matrix

    m = get_majorana_operators(d)

    majorana_string = reduce(np.dot, [m[i] for i in indices])

    assert np.isclose(
        state.get_majorana_monomial_expectation_value(indices),
        np.trace(density_matrix @ majorana_string),
    )


@pytest.mark.monkey
@for_all_connectors
def test_get_majorana_monomial_expectation_value_empty(
    connector, generate_fermionic_gaussian_hamiltonian
):
    d = 3

    parent_hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    program = pq.Program(
        [pq.fermionic.ParentHamiltonian(hamiltonian=parent_hamiltonian)]
    )
    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)
    state = simulator.execute(program).state

    density_matrix = state.density_matrix

    empty_indices = np.array([], dtype=int)

    assert np.isclose(
        state.get_majorana_monomial_expectation_value(empty_indices),
        np.trace(density_matrix),
    )


@pytest.mark.monkey
@for_all_connectors
def test_get_majorana_monomial_expectation_value_random_with_possible_multiplicities(
    connector, get_majorana_operators, generate_fermionic_gaussian_hamiltonian
):
    d = 3

    parent_hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    program = pq.Program(
        [pq.fermionic.ParentHamiltonian(hamiltonian=parent_hamiltonian)]
    )
    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)
    state = simulator.execute(program).state
    initial_length = np.random.randint(1, 10)

    indices = np.random.randint(0, 2 * d, initial_length)

    density_matrix = state.density_matrix

    m = get_majorana_operators(d)

    majorana_string = reduce(np.dot, [m[i] for i in indices])

    assert np.isclose(
        state.get_majorana_monomial_expectation_value(indices),
        np.trace(density_matrix @ majorana_string),
    )


@pytest.mark.monkey
@for_all_connectors
def test_reduced_state_correlation_matrix_random(
    connector,
    generate_fermionic_gaussian_hamiltonian,
):
    d = 3

    hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state
    reduced_state = state.reduced(modes=(0, 1))

    gamma = reduced_state.correlation_matrix
    dr = d - 1

    gamma11 = gamma[:dr, :dr]
    gamma12 = gamma[:dr, dr:]
    gamma21 = gamma[dr:, :dr]
    gamma22 = gamma[dr:, dr:]

    assert is_selfadjoint(gamma11)
    assert is_selfadjoint(gamma22)

    assert is_skew_symmetric(gamma12)
    assert is_skew_symmetric(gamma21)

    assert np.allclose(gamma11, np.identity(dr) - gamma22.conj())
    assert np.allclose(gamma12, -gamma21.conj())


@for_all_connectors
def test_reduced_state_correlation_matrix(connector):
    d = 3

    A = np.array([[1, 2j, 3j], [-2j, 5, 6], [-3j, 6, 7]])
    B = np.array([[0, 1j, 2], [-1j, 0, 3], [-2, -3, 0]])

    hamiltonian = np.block([[-A.conj(), B], [-B.conj(), A]])

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    reduced_state = state.reduced(modes=(0, 1))
    reduced_state.validate()

    assert np.allclose(
        reduced_state.correlation_matrix,
        np.array(
            [
                [
                    0.33416157 - 0.0j,
                    -0.0401874 + 0.01146957j,
                    0.0 + 0.0j,
                    -0.19852528 + 0.31881949j,
                ],
                [
                    -0.0401874 - 0.01146957j,
                    0.23155699 - 0.0j,
                    0.19852528 - 0.31881949j,
                    0.0 + 0.0j,
                ],
                [
                    -0.0 + 0.0j,
                    0.19852528 + 0.31881949j,
                    0.66583843 - 0.0j,
                    0.0401874 + 0.01146957j,
                ],
                [
                    -0.19852528 - 0.31881949j,
                    -0.0 + 0.0j,
                    0.0401874 - 0.01146957j,
                    0.76844301 - 0.0j,
                ],
            ]
        ),
    )


@pytest.mark.monkey
@for_all_connectors
def test_vacuum_evolved_with_passive_GaussianHamiltonian_stays_vacuum(
    connector,
    generate_passive_fermionic_gaussian_hamiltonian,
):
    d = 3

    hamiltonian = generate_passive_fermionic_gaussian_hamiltonian(d)

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    with pq.Program() as vacuum_program:
        pq.Q() | pq.Vacuum()

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state
    vacuum_state = simulator.execute(vacuum_program).state

    assert state == vacuum_state


@pytest.mark.monkey
@for_all_connectors
def test_correlation_matrix_ParentHamiltonian(
    connector, generate_fermionic_gaussian_hamiltonian
):
    d = 3

    parent_hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    program = pq.Program(
        [pq.fermionic.ParentHamiltonian(hamiltonian=parent_hamiltonian)]
    )
    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)
    state = simulator.execute(program).state
    state.validate()

    assert np.allclose(
        state.correlation_matrix,
        np.linalg.inv(np.identity(2 * d) + connector.expm(2 * parent_hamiltonian)),
    )


@for_all_connectors
def test_get_parent_hamiltonian_Vacuum(connector):
    d = 3

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    with pytest.raises(pq.api.exceptions.PiquassoException) as error:
        state.get_parent_hamiltonian()

    error_message = error.value.args[0]

    assert error_message == (
        "Cannot calculate parent Hamiltonian, since the correlation matrix is "
        "singular."
    )


@pytest.mark.monkey
@for_all_connectors
def test_parent_hamiltonian_roundtrip(
    connector, generate_fermionic_gaussian_hamiltonian
):
    d = 3

    parent_hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    program = pq.Program(
        [pq.fermionic.ParentHamiltonian(hamiltonian=parent_hamiltonian)]
    )
    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)
    state = simulator.execute(program).state
    state.validate()

    assert np.allclose(parent_hamiltonian, state.get_parent_hamiltonian())


@for_all_connectors
def test_ParentHamiltonian_invalid_hamiltonian_raises_InvalidState(connector):
    d = 1

    parent_hamiltonian = np.array([[1, 2], [3, 4]])

    program = pq.Program(
        [pq.fermionic.ParentHamiltonian(hamiltonian=parent_hamiltonian)]
    )
    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)
    with pytest.raises(pq.api.exceptions.InvalidParameter) as error:
        simulator.execute(program).state

    error_message = error.value.args[0]

    assert error_message == "Invalid Hamiltonian specified."


@for_all_connectors
def test_density_matrix_single_mode(connector):
    eps = 1 / 3

    parent_hamiltonian = np.array([[eps, 0], [0, -eps]])

    program = pq.Program(
        [pq.fermionic.ParentHamiltonian(hamiltonian=parent_hamiltonian)]
    )
    simulator = pq.fermionic.GaussianSimulator(d=1, connector=connector)
    state = simulator.execute(program).state
    state.validate()

    density_matrix = state.density_matrix

    assert is_selfadjoint(density_matrix)
    eigvals = np.linalg.eigvals(density_matrix)

    assert all_in_interval(eigvals, 0, 1)

    assert np.isclose(np.trace(density_matrix), 1.0)

    f = 1 / (1 + np.exp(2 * eps))

    assert np.allclose(density_matrix, np.array([[1 - f, 0], [0, f]]))


@for_all_connectors
def test_density_matrix_two_mode_simple(connector):
    eps1 = 1 / 3
    eps2 = 1 / 4

    eps = np.array([eps1, eps2])

    parent_hamiltonian = np.diag(np.concatenate([eps, -eps]))

    program = pq.Program(
        [pq.fermionic.ParentHamiltonian(hamiltonian=parent_hamiltonian)]
    )
    simulator = pq.fermionic.GaussianSimulator(d=2, connector=connector)
    state = simulator.execute(program).state
    state.validate()

    density_matrix = state.density_matrix

    assert is_selfadjoint(density_matrix)
    eigvals = np.linalg.eigvals(density_matrix)

    assert all_in_interval(eigvals, 0, 1)

    assert np.isclose(np.trace(density_matrix), 1.0)

    f = 1 / (1 + np.exp(2 * eps))

    from piquasso.fermionic._utils import tensor_product

    single_mode_dms = []

    for fi in f:
        single_mode_dms.append(np.diag([1 - fi, fi]))

    assert np.allclose(density_matrix, tensor_product(single_mode_dms))


@for_all_connectors
@pytest.mark.monkey
def test_density_matrix_2_mode(connector, generate_fermionic_gaussian_hamiltonian):
    d = 2

    parent_hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    program = pq.Program(
        [pq.fermionic.ParentHamiltonian(hamiltonian=parent_hamiltonian)]
    )
    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)
    state = simulator.execute(program).state
    state.validate()

    density_matrix = state.density_matrix

    assert is_selfadjoint(density_matrix)
    eigvals = np.linalg.eigvals(density_matrix)

    assert all_in_interval(eigvals, 0, 1)

    assert np.isclose(np.trace(density_matrix), 1.0)

    bigH = pq.fermionic._utils.get_fermionic_hamiltonian(parent_hamiltonian, connector)

    assert np.allclose(
        density_matrix, connector.expm(bigH) / np.trace(connector.expm(bigH))
    )


@pytest.mark.monkey
@for_all_connectors
def test_density_matrix(connector, generate_fermionic_gaussian_hamiltonian):
    d = 3

    parent_hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    program = pq.Program(
        [pq.fermionic.ParentHamiltonian(hamiltonian=parent_hamiltonian)]
    )
    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)
    state = simulator.execute(program).state
    state.validate()

    density_matrix = state.density_matrix

    assert is_selfadjoint(density_matrix)
    eigvals = np.linalg.eigvals(density_matrix)

    assert all_in_interval(eigvals, 0, 1)

    assert np.isclose(np.trace(density_matrix), 1.0)


@for_all_connectors
def test_identical_state_vector_overlap(connector):
    d = 3

    state_vector = [1, 0, 1]

    with pq.Program() as program:
        pq.Q() | pq.StateVector(state_vector)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    assert np.isclose(state.overlap(state), 1.0)


@for_all_connectors
def test_different_state_vector_overlap(connector):
    d = 3

    state_vector_1 = [1, 0, 1]

    state_vector_2 = [0, 1, 1]

    with pq.Program() as program_1:
        pq.Q() | pq.StateVector(state_vector_1)

    with pq.Program() as program_2:
        pq.Q() | pq.StateVector(state_vector_2)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state_1 = simulator.execute(program_1).state
    state_2 = simulator.execute(program_2).state

    assert np.isclose(state_1.overlap(state_2), 0.0)
    assert np.isclose(state_2.overlap(state_1), 0.0)


@pytest.mark.monkey
@for_all_connectors
def test_covariance_matrix_GaussianHamiltonian_equivalence_from_Vacuum(
    connector, generate_fermionic_gaussian_hamiltonian
):
    d = 3

    gate_hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    with pq.Program() as preparation:
        pq.Q() | pq.Vacuum()

    with pq.Program() as program:
        pq.Q() | preparation
        pq.Q() | pq.fermionic.GaussianHamiltonian(gate_hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    initial_state = simulator.execute(preparation).state
    final_state = simulator.execute(program).state

    omega = get_omega(d)

    gate_hamiltonian_majorana = -1j * omega @ gate_hamiltonian @ omega.conj().T

    indices = xxpp_to_xpxp_indices(d)

    covariance_unitary = connector.expm(-2 * gate_hamiltonian_majorana)[
        np.ix_(indices, indices)
    ]

    assert np.allclose(
        final_state.covariance_matrix,
        covariance_unitary
        @ initial_state.covariance_matrix
        @ covariance_unitary.conj().T,
    )


@pytest.mark.monkey
@for_all_connectors
def test_correlation_matrix_GaussianHamiltonian_equivalence_from_Vacuum(
    connector, generate_fermionic_gaussian_hamiltonian
):
    d = 3

    gate_hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    with pq.Program() as preparation:
        pq.Q() | pq.Vacuum()

    with pq.Program() as program:
        pq.Q() | preparation
        pq.Q() | pq.fermionic.GaussianHamiltonian(gate_hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    initial_state = simulator.execute(preparation).state
    final_state = simulator.execute(program).state

    correlation_unitary = connector.expm(2j * gate_hamiltonian)

    assert np.allclose(
        final_state.correlation_matrix,
        correlation_unitary
        @ initial_state.correlation_matrix
        @ correlation_unitary.conj().T,
    )


@pytest.mark.monkey
@for_all_connectors
def test_density_matrix_GaussianHamiltonian_equivalence_from_Vacuum(
    connector, generate_fermionic_gaussian_hamiltonian
):
    d = 3

    gate_hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    with pq.Program() as preparation:
        pq.Q() | pq.Vacuum()

    with pq.Program() as program:
        pq.Q() | preparation
        pq.Q() | pq.fermionic.GaussianHamiltonian(gate_hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    initial_state = simulator.execute(preparation).state
    final_state = simulator.execute(program).state

    bigH = pq.fermionic._utils.get_fermionic_hamiltonian(gate_hamiltonian, connector)

    gate_unitary = connector.expm(1j * bigH)

    assert np.allclose(
        final_state.density_matrix,
        gate_unitary @ initial_state.density_matrix @ gate_unitary.conj().T,
    )


@pytest.mark.monkey
@for_all_connectors
def test_density_matrix_GaussianHamiltonian_equivalence_1_mode(
    connector, generate_fermionic_gaussian_hamiltonian
):
    d = 1

    parent_hamiltonian = generate_fermionic_gaussian_hamiltonian(d)
    gate_hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    program = pq.Program(
        [pq.fermionic.ParentHamiltonian(hamiltonian=parent_hamiltonian)]
    )
    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)
    initial_state = simulator.execute(program).state

    with pq.Program() as program:
        pq.Q() | pq.fermionic.GaussianHamiltonian(gate_hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    final_state = simulator.execute(program, initial_state=initial_state).state

    bigH = pq.fermionic._utils.get_fermionic_hamiltonian(gate_hamiltonian, connector)

    gate_unitary = connector.expm(1j * bigH)

    assert np.allclose(
        final_state.density_matrix,
        gate_unitary @ initial_state.density_matrix @ gate_unitary.conj().T,
    )


@pytest.mark.monkey
@for_all_connectors
def test_density_matrix_GaussianHamiltonian_equivalence_2_mode(
    connector, generate_fermionic_gaussian_hamiltonian
):
    d = 2

    parent_hamiltonian = generate_fermionic_gaussian_hamiltonian(d)
    gate_hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    program = pq.Program(
        [pq.fermionic.ParentHamiltonian(hamiltonian=parent_hamiltonian)]
    )
    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)
    initial_state = simulator.execute(program).state

    with pq.Program() as program:
        pq.Q() | pq.fermionic.GaussianHamiltonian(gate_hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    final_state = simulator.execute(program, initial_state=initial_state).state

    bigH = pq.fermionic._utils.get_fermionic_hamiltonian(gate_hamiltonian, connector)

    gate_unitary = connector.expm(1j * bigH)

    assert np.allclose(
        final_state.density_matrix,
        gate_unitary @ initial_state.density_matrix @ gate_unitary.conj().T,
    )


@pytest.mark.monkey
@for_all_connectors
def test_density_matrix_GaussianHamiltonian_equivalence(
    connector, generate_fermionic_gaussian_hamiltonian
):
    d = 3

    parent_hamiltonian = generate_fermionic_gaussian_hamiltonian(d)
    gate_hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    program = pq.Program(
        [pq.fermionic.ParentHamiltonian(hamiltonian=parent_hamiltonian)]
    )
    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)
    initial_state = simulator.execute(program).state

    with pq.Program() as program:
        pq.Q() | pq.fermionic.GaussianHamiltonian(gate_hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    final_state = simulator.execute(program, initial_state=initial_state).state

    bigH = pq.fermionic._utils.get_fermionic_hamiltonian(gate_hamiltonian, connector)

    gate_unitary = connector.expm(1j * bigH)

    assert np.allclose(
        final_state.density_matrix,
        gate_unitary @ initial_state.density_matrix @ gate_unitary.conj().T,
    )


@for_all_connectors
def test_Vacuum_density_matrix(connector):
    d = 2

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

    simulator = pq.fermionic.GaussianSimulator(d=d)

    state = simulator.execute(program).state

    assert np.allclose(state.density_matrix, np.diag([1, 0, 0, 0]))


@for_all_connectors
def test_StateVector_density_matrix(connector):
    d = 2

    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 0])

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    assert np.allclose(state.density_matrix, np.diag([0, 0, 1, 0]))


@for_all_connectors
def test_parity_StateVector(connector):
    d = 3

    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 1, 0])

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    parity = state.get_parity_operator_expectation_value()

    assert np.isclose(parity, -1)


@pytest.mark.monkey
@for_all_connectors
def test_parity_StateVector_random(connector):
    d = np.random.randint(1, 10)

    state_vector = np.random.randint(0, 2, d)

    with pq.Program() as program:
        pq.Q() | pq.StateVector(state_vector)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    parity = state.get_parity_operator_expectation_value()

    assert np.isclose(parity, (-1) ** np.sum(state_vector))


@pytest.mark.monkey
@for_all_connectors
def test_parity_is_invariant_under_linear_transformations(
    connector,
    generate_fermionic_gaussian_hamiltonian,
):
    d = np.random.randint(1, 10)

    state_vector = np.random.randint(0, 2, d)

    H = generate_fermionic_gaussian_hamiltonian(d)

    with pq.Program() as program:
        pq.Q() | pq.StateVector(state_vector)

        pq.Q() | pq.fermionic.GaussianHamiltonian(H)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    parity = state.get_parity_operator_expectation_value()

    assert np.isclose(parity, (-1) ** np.sum(state_vector))


@pytest.mark.monkey
@for_all_connectors
def test_overlap_Interferometer(connector, generate_unitary_matrix):
    d = 3

    U1 = generate_unitary_matrix(d)
    U2 = generate_unitary_matrix(d)

    state_vector = [1, 0, 1]

    with pq.Program() as program1:
        pq.Q() | pq.StateVector(state_vector)
        pq.Q() | pq.Interferometer(U1)

    with pq.Program() as program2:
        pq.Q() | pq.StateVector(state_vector)
        pq.Q() | pq.Interferometer(U2)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state1 = simulator.execute(program1).state
    state2 = simulator.execute(program2).state

    state1.validate()
    state2.validate()

    expected_overlap = np.trace(state1.density_matrix @ state2.density_matrix)

    assert np.isclose(state1.overlap(state2), expected_overlap)


@pytest.mark.monkey
@for_all_connectors
def test_overlap(connector, generate_fermionic_gaussian_hamiltonian):
    d = 3

    parent_hamiltonian_1 = generate_fermionic_gaussian_hamiltonian(d)
    parent_hamiltonian_2 = generate_fermionic_gaussian_hamiltonian(d)

    program = pq.Program(
        [pq.fermionic.ParentHamiltonian(hamiltonian=parent_hamiltonian_1)]
    )
    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)
    state1 = simulator.execute(program).state

    program = pq.Program(
        [pq.fermionic.ParentHamiltonian(hamiltonian=parent_hamiltonian_2)]
    )
    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)
    state2 = simulator.execute(program).state

    state1.validate()
    state2.validate()

    expected_overlap = np.trace(state1.density_matrix @ state2.density_matrix)

    assert np.isclose(state1.overlap(state2), expected_overlap)


@pytest.mark.monkey
@for_all_connectors
def test_fock_probabilities_density_matrix_equivalence(
    connector, generate_fermionic_gaussian_hamiltonian
):
    d = 3
    parent_hamiltonian = generate_fermionic_gaussian_hamiltonian(d)

    program = pq.Program(
        [pq.fermionic.ParentHamiltonian(hamiltonian=parent_hamiltonian)]
    )
    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)
    state = simulator.execute(program).state
    state.validate()

    fock_probabilities_from_density_matrix = np.real(np.diag(state.density_matrix))

    assert np.allclose(state.fock_probabilities, fock_probabilities_from_density_matrix)


@for_all_connectors
def test_density_matrix_Interferometer_2_by_2_simple(connector):
    d = 2
    theta = np.pi / 5
    phi = np.pi / 7

    U = np.array(
        [
            [np.cos(theta), -np.exp(-1j * phi) * np.sin(theta)],
            [np.exp(1j * phi) * np.sin(theta), np.cos(theta)],
        ]
    )

    with pq.Program() as preparation:
        pq.Q(0, 1) | pq.StateVector([0, 1])

    with pq.Program() as program:
        pq.Q(0, 1) | preparation
        pq.Q(0, 1) | pq.Interferometer(U)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)
    initial_state = simulator.execute(preparation).state
    final_state = simulator.execute(program).state

    # The basis is ordered as [0, 0], [0, 1], [1, 0], [1, 1], but the unitary acting
    # on the one-particle Hilbert space is understood in the basis [1, 0], [0, 1],
    # therefore we have to flip it.
    indices = binary_to_fock_indices(d)

    initial_state_density_matrix = initial_state.density_matrix[
        np.ix_(indices, indices)
    ]
    final_state_density_matrix = final_state.density_matrix[np.ix_(indices, indices)]

    fock_space_unitary = block_diag(np.array([1.0]), U, np.array([1.0]))

    assert np.allclose(
        final_state_density_matrix,
        fock_space_unitary @ initial_state_density_matrix @ fock_space_unitary.T.conj(),
    )


@pytest.mark.monkey
@for_all_connectors
def test_density_matrix_Interferometer_2_by_2_random(
    connector, generate_unitary_matrix
):
    d = 2
    U = generate_unitary_matrix(d)

    with pq.Program() as preparation:
        pq.Q(0, 1) | pq.StateVector([0, 1])

    with pq.Program() as program:
        pq.Q(0, 1) | preparation
        pq.Q(0, 1) | pq.Interferometer(U)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)
    initial_state = simulator.execute(preparation).state
    final_state = simulator.execute(program).state

    # The basis is ordered as [0, 0], [0, 1], [1, 0], [1, 1], but the unitary acting
    # on the one-particle Hilbert space is understood in the basis [1, 0], [0, 1],
    # therefore we have to flip it.
    indices = binary_to_fock_indices(d)

    initial_state_density_matrix = initial_state.density_matrix[
        np.ix_(indices, indices)
    ]
    final_state_density_matrix = final_state.density_matrix[np.ix_(indices, indices)]

    fock_space_unitary = block_diag(np.array([1.0]), U, np.array([1.0]))

    assert np.allclose(
        final_state_density_matrix,
        fock_space_unitary @ initial_state_density_matrix @ fock_space_unitary.T.conj(),
    )


@for_all_connectors
def test_density_matrix_StateVector_ordering(connector):
    d = 3

    state_vectors = [
        [0, 0, 0],
        [0, 0, 1],
        [0, 1, 0],
        [0, 1, 1],
        [1, 0, 0],
        [1, 0, 1],
        [1, 1, 0],
        [1, 1, 1],
    ]

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    for i in range(len(state_vectors)):
        with pq.Program() as preparation:
            pq.Q(0, 1, 2) | pq.StateVector(state_vectors[i])

        state = simulator.execute(preparation).state

        density_matrix = state.density_matrix

        assert np.isclose(density_matrix[i, i], 1.0)


@pytest.mark.monkey
@for_all_connectors
def test_density_matrix_Interferometer_3_by_3_random(
    connector, generate_unitary_matrix
):
    d = 3

    U = generate_unitary_matrix(d)

    with pq.Program() as preparation:
        pq.Q(0, 1, 2) | pq.StateVector([1, 0, 0])

    with pq.Program() as program:
        pq.Q(0, 1, 2) | preparation
        pq.Q(0, 1, 2) | pq.Interferometer(U)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)
    initial_state = simulator.execute(preparation).state
    final_state = simulator.execute(program).state

    # The basis is ordered as lexicographically, but the unitary acting
    # on the one-particle Hilbert space is understood in a different basis,
    # therefore we have to flip it.
    indices = binary_to_fock_indices(d)

    initial_state_density_matrix = initial_state.density_matrix[
        np.ix_(indices, indices)
    ]
    final_state_density_matrix = final_state.density_matrix[np.ix_(indices, indices)]

    does_not_matter = np.random.rand(d, d)

    fock_space_unitary = block_diag(
        np.array([1.0]), U, does_not_matter, np.array([1.0])
    )

    assert np.allclose(
        final_state_density_matrix,
        fock_space_unitary @ initial_state_density_matrix @ fock_space_unitary.T.conj(),
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest

from scipy.stats import unitary_group

import piquasso as pq
from piquasso.api.exceptions import InvalidParameter


@pytest.fixture
def interferometer_matrix():
    return np.array(
        [
            [0, 0, 1, 0, 0],
            [0, 0, 0, 1, 0],
            [0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0],
            [0, 1, 0, 0, 0],
        ],
    )


def test_sampling_raises_InvalidParameter_for_negative_shot_value(
    interferometer_matrix,
):
    invalid_shots = -1

    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1, 1, 0, 0])
        pq.Q() | pq.Interferometer(interferometer_matrix)
        pq.Q() | pq.ParticleNumberMeasurement()

    simulator = pq.SamplingSimulator(d=5)

    with pytest.raises(InvalidParameter):
        simulator.execute(program, invalid_shots)


def test_sampling_samples_number(interferometer_matrix):
    shots = 100

    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1, 1, 0, 0])
        pq.Q() | pq.Interferometer(interferometer_matrix)
        pq.Q() | pq.ParticleNumberMeasurement()

    simulator = pq.SamplingSimulator(d=5)
    result = simulator.execute(program, shots)

    assert len(result.samples) == shots, (
        f"Expected {shots} samples, " f"got: {len(program.result)}"
    )


def test_sampling_mode_permutation(interferometer_matrix):
    shots = 1

    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1, 1, 0, 0])
        pq.Q() | pq.Interferometer(interferometer_matrix)
        pq.Q() | pq.ParticleNumberMeasurement()

    simulator = pq.SamplingSimulator(d=5)
    result = simulator.execute(program, shots)

    sample = result.samples[0]
    assert np.allclose(
        sample, [1, 0, 0, 1, 1]
    ), f"Expected [1, 0, 0, 1, 1], got: {sample}"


def test_sampling_multiple_samples_for_permutation_interferometer(
    interferometer_matrix,
):
    shots = 2

    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1, 1, 0, 0])
        pq.Q() | pq.Interferometer(interferometer_matrix)
        pq.Q() | pq.ParticleNumberMeasurement()

    simulator = pq.SamplingSimulator(d=5)
    result = simulator.execute(program, shots)

    samples = result.samples
    first_sample = samples[0]
    second_sample = samples[1]

    assert np.allclose(
        first_sample, second_sample
    ), f"Expected same samples, got: {first_sample} & {second_sample}"


basis_states = [
    (0, 1),
    (1, 0),
    (2, 0),
]


@pytest.mark.parametrize("input_state", basis_states)
def test_counts(input_state):
    shots = 100

    expected = {input_state: shots}
    with pq.Program() as program:
        pq.Q() | pq.StateVector(input_state)
        pq.Q() | pq.ParticleNumberMeasurement()

    simulator = pq.SamplingSimulator(d=2)
    res = simulator.execute(program, shots=shots)
    counts = res.get_counts()
    assert len(counts) == len(expected)
    for r, e in zip(counts.items(), expected.items()):
        assert r[0] == e[0]
        assert np.isclose(r[1] / shots, e[1] / shots)


@pytest.mark.parametrize(
    "simulator, measurement_class",
    [
        (pq.PureFockSimulator, pq.HomodyneMeasurement),
        (pq.GaussianSimulator, pq.HeterodyneMeasurement),
    ],
)
def test_counts_raises(simulator, measurement_class):
    shots = 100

    with pq.Program() as program:
        if simulator == pq.PureFockSimulator:
            pq.Q() | pq.StateVector((0, 1, 0))
        else:
            pq.Q(0) | pq.Fourier()
            pq.Q(0, 2) | pq.Squeezing2(r=0.5, phi=0)
            pq.Q(0, 1) | pq.Beamsplitter(theta=1, phi=0)
        pq.Q() | measurement_class()

    simulator = simulator(d=3)
    res = simulator.execute(program, shots=shots)
    with pytest.raises(
        NotImplementedError, match="method only supports samples that contain integers"
    ):
        res.get_counts()


def test_mach_zehnder():
    int_ = np.pi / 3
    ext = np.pi / 4

    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1, 1, 0, 0])
        pq.Q(0, 1) | pq.MachZehnder(int_=int_, ext=ext)
        pq.Q() | pq.ParticleNumberMeasurement()

    simulator = pq.SamplingSimulator(d=5)
    simulator.execute(program, shots=1)


def test_fourier():
    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1, 1, 0, 0])
        pq.Q(0) | pq.Fourier()
        pq.Q() | pq.ParticleNumberMeasurement()

    simulator = pq.SamplingSimulator(d=5)
    simulator.execute(program, shots=1)


def test_multiple_StateVector_with_ParticleNumberMeasurement_raises_error():
    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1, 1, 0, 0]) * 1 / np.sqrt(2)
        pq.Q() | pq.StateVector([1, 1, 0, 1, 0]) * 1 / np.sqrt(2)

        pq.Q(0) | pq.Fourier()
        pq.Q() | pq.ParticleNumberMeasurement()

    simulator = pq.SamplingSimulator(d=5)

    with pytest.raises(pq.api.exceptions.NotImplementedCalculation) as error:
        simulator.execute(program, shots=1)

    assert error.value.args[0] == (
        "The instruction ParticleNumberMeasurement(modes=(0, 1, 2, 3, 4)) is "
        "not supported for states defined using multiple 'StateVector' instructions.\n"
        "If you need this feature to be implemented, please create an issue at "
        "https://github.com/Budapest-Quantum-Computing-Group/piquasso/issues"
    )


def test_uniform_loss():
    d = 5

    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1, 1, 0, 0])

        for i in range(d):
            pq.Q(i) | pq.Loss(transmissivity=0.9)

        pq.Q() | pq.ParticleNumberMeasurement()

    simulator = pq.SamplingSimulator(d=d)
    state = simulator.execute(program, shots=1).state

    assert state.is_lossy


def test_general_loss():
    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1, 1, 0, 0])

        pq.Q(0) | pq.Loss(transmissivity=0.4)
        pq.Q(1) | pq.Loss(transmissivity=0.5)

        pq.Q() | pq.ParticleNumberMeasurement()

    simulator = pq.SamplingSimulator(d=5)
    simulator.execute(program, shots=1)


def test_boson_sampling_seeded():
    seed_sequence = 123

    U = np.array(
        [
            [
                -0.17959207 - 0.29175972j,
                -0.64550941 - 0.02897055j,
                -0.023922 - 0.38854671j,
                -0.07908932 + 0.35617293j,
                -0.39779191 + 0.14902272j,
            ],
            [
                -0.36896208 + 0.19468375j,
                -0.11545557 + 0.20434514j,
                0.25548079 + 0.05220164j,
                -0.51002161 + 0.38442256j,
                0.48106678 - 0.25210091j,
            ],
            [
                0.25912844 + 0.16131742j,
                0.11886251 + 0.12632645j,
                0.69028213 - 0.25734432j,
                0.01276639 + 0.05841739j,
                0.03713264 + 0.57364845j,
            ],
            [
                -0.20314019 - 0.18973473j,
                0.59146854 + 0.28605532j,
                -0.11096495 - 0.26870144j,
                -0.47290354 - 0.0489408j,
                -0.42459838 + 0.01554643j,
            ],
            [
                -0.5021973 - 0.53474291j,
                0.24524545 - 0.0741398j,
                0.37786104 + 0.10225255j,
                0.46696955 + 0.10636677j,
                0.07171789 - 0.09194236j,
            ],
        ]
    )

    with pq.Program() as program:
        pq.Q(all) | pq.StateVector([2, 1, 1, 0, 1])

        pq.Q(all) | pq.Interferometer(U)

        pq.Q(all) | pq.ParticleNumberMeasurement()

    simulator = pq.SamplingSimulator(d=5, config=pq.Config(seed_sequence=seed_sequence))
    samples = simulator.execute(program, shots=20).samples

    expected_samples = [
        (3, 0, 0, 0, 2),
        (0, 0, 1, 2, 2),
        (0, 1, 0, 4, 0),
        (1, 0, 0, 0, 4),
        (0, 1, 0, 1, 3),
        (1, 0, 0, 3, 1),
        (2, 0, 3, 0, 0),
        (3, 1, 1, 0, 0),
        (0, 1, 0, 3, 1),
        (1, 2, 0, 0, 2),
        (1, 0, 4, 0, 0),
        (0, 1, 1, 2, 1),
        (0, 0, 1, 1, 3),
        (0, 2, 0, 1, 2),
        (2, 0, 1, 0, 2),
        (0, 0, 3, 1, 1),
        (0, 0, 4, 1, 0),
        (2, 0, 0, 0, 3),
        (0, 0, 2, 1, 2),
        (1, 1, 0, 1, 2),
    ]

    assert samples == expected_samples


def test_LossyInterferometer_boson_sampling_seeded():
    seed_sequence = 123

    U = np.array(
        [
            [
                -0.17959207 - 0.29175972j,
                -0.64550941 - 0.02897055j,
                -0.023922 - 0.38854671j,
                -0.07908932 + 0.35617293j,
                -0.39779191 + 0.14902272j,
            ],
            [
                -0.36896208 + 0.19468375j,
                -0.11545557 + 0.20434514j,
                0.25548079 + 0.05220164j,
                -0.51002161 + 0.38442256j,
                0.48106678 - 0.25210091j,
            ],
            [
                0.25912844 + 0.16131742j,
                0.11886251 + 0.12632645j,
                0.69028213 - 0.25734432j,
                0.01276639 + 0.05841739j,
                0.03713264 + 0.57364845j,
            ],
            [
                -0.20314019 - 0.18973473j,
                0.59146854 + 0.28605532j,
                -0.11096495 - 0.26870144j,
                -0.47290354 - 0.0489408j,
                -0.42459838 + 0.01554643j,
            ],
            [
                -0.5021973 - 0.53474291j,
                0.24524545 - 0.0741398j,
                0.37786104 + 0.10225255j,
                0.46696955 + 0.10636677j,
                0.07171789 - 0.09194236j,
            ],
        ]
    )

    singular_values = np.array([0.9, 0.8, 0.7, 0.6, 0.5])

    lossy_interferometer_matrix = U @ np.diag(singular_values) @ U @ U.T

    with pq.Program() as program:
        pq.Q(all) | pq.StateVector([2, 1, 1, 0, 1])

        pq.Q(all) | pq.LossyInterferometer(lossy_interferometer_matrix)

        pq.Q(all) | pq.ParticleNumberMeasurement()

    simulator = pq.SamplingSimulator(d=5, config=pq.Config(seed_sequence=seed_sequence))
    samples = simulator.execute(program, shots=20).samples

    expected_samples = [
        (0, 0, 3, 0, 0),
        (0, 0, 1, 1, 0),
        (0, 0, 1, 1, 0),
        (1, 0, 0, 1, 0),
        (0, 0, 1, 0, 0),
        (0, 0, 1, 0, 0),
        (1, 1, 1, 0, 0),
        (0, 2, 0, 0, 2),
        (0, 0, 1, 0, 0),
        (1, 0, 1, 0, 2),
        (1, 0, 0, 0, 3),
        (0, 0, 1, 1, 2),
        (0, 0, 0, 1, 0),
        (0, 0, 0, 1, 2),
        (0, 1, 1, 1, 0),
        (0, 1, 2, 0, 0),
        (0, 1, 0, 0, 1),
        (1, 0, 0, 1, 1),
        (0, 0, 2, 1, 0),
        (0, 1, 1, 0, 0),
    ]

    assert samples == expected_samples


def test_LossyInterferometer_boson_sampling_uniform_losses():
    seed_sequence = 123

    U = np.array(
        [
            [
                -0.17959207 - 0.29175972j,
                -0.64550941 - 0.02897055j,
                -0.023922 - 0.38854671j,
                -0.07908932 + 0.35617293j,
                -0.39779191 + 0.14902272j,
            ],
            [
                -0.36896208 + 0.19468375j,
                -0.11545557 + 0.20434514j,
                0.25548079 + 0.05220164j,
                -0.51002161 + 0.38442256j,
                0.48106678 - 0.25210091j,
            ],
            [
                0.25912844 + 0.16131742j,
                0.11886251 + 0.12632645j,
                0.69028213 - 0.25734432j,
                0.01276639 + 0.05841739j,
                0.03713264 + 0.57364845j,
            ],
            [
                -0.20314019 - 0.18973473j,
                0.59146854 + 0.28605532j,
                -0.11096495 - 0.26870144j,
                -0.47290354 - 0.0489408j,
                -0.42459838 + 0.01554643j,
            ],
            [
                -0.5021973 - 0.53474291j,
                0.24524545 - 0.0741398j,
                0.37786104 + 0.10225255j,
                0.46696955 + 0.10636677j,
                0.07171789 - 0.09194236j,
            ],
        ]
    )

    singular_values = np.array([0.9] * 5)

    lossy_interferometer_matrix = U @ np.diag(singular_values) @ U @ U.T

    with pq.Program() as program:
        pq.Q(all) | pq.StateVector([2, 1, 1, 0, 1])

        pq.Q(all) | pq.LossyInterferometer(lossy_interferometer_matrix)

        pq.Q(all) | pq.ParticleNumberMeasurement()

    simulator = pq.SamplingSimulator(d=5, config=pq.Config(seed_sequence=seed_sequence))
    samples = simulator.execute(program, shots=20).samples

    expected_samples = [
        (1, 0, 2, 0, 0),
        (1, 0, 0, 1, 1),
        (0, 2, 1, 0, 2),
        (1, 0, 0, 0, 2),
        (5, 0, 0, 0, 0),
        (0, 2, 2, 0, 1),
        (0, 3, 2, 0, 0),
        (0, 1, 1, 2, 1),
        (1, 1, 0, 0, 3),
        (0, 0, 0, 2, 2),
        (0, 2, 1, 0, 2),
        (2, 0, 2, 0, 0),
        (0, 1, 0, 0, 3),
        (0, 1, 3, 0, 0),
        (1, 1, 0, 0, 2),
        (0, 0, 4, 1, 0),
        (1, 3, 0, 0, 1),
        (2, 1, 1, 0, 1),
        (0, 2, 1, 0, 2),
        (0, 1, 1, 0, 2),
    ]

    assert samples == expected_samples


@pytest.mark.monkey
def test_post_select_random_unitary():
    d = 3

    interferometer_matrix = unitary_group.rvs(d)

    postselect_modes = (1, 2)

    with pq.Program() as program:
        pq.Q(all) | pq.StateVector([2, 1, 0])

        pq.Q(all) | pq.Interferometer(interferometer_matrix)

        pq.Q(all) | pq.PostSelectPhotons(
            postselect_modes=postselect_modes, photon_counts=(1, 0)
        )

    simulator = pq.SamplingSimulator(d=d, config=pq.Config(cutoff=4))

    state = simulator.execute(program).state

    state.normalize()
    state.validate()

    assert state.d == d - len(postselect_modes)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

import numba as nb


@nb.njit(cache=True)
def _get_inverse_permutation(p):
    s = np.empty(p.size, dtype=np.int32)
    for i in np.arange(p.size):
        s[p[i]] = i

    return s


@nb.njit(cache=True)
def match_occupation_numbers(nvec_orig):
    dtype = nvec_orig.dtype
    nvec = np.copy(nvec_orig)
    edge_reps = []
    edge_indices = []

    if len(nvec_orig) == 1:
        # NOTE: This function only executed when `sum(nvec_orig) % 2 == 0`.
        return np.array([nvec_orig[0] // 2], dtype=dtype), np.array([0, 0], dtype=dtype)

    while sum(nvec) > 1:
        sorter = np.argsort(nvec)
        nvec_sorted = nvec[sorter]

        n0_over_2 = nvec_sorted[-1] // 2
        n1 = nvec_sorted[-2]

        if n0_over_2 > nvec_sorted[-2]:
            nvec_sorted[-1] -= 2 * n0_over_2
            edge_reps.append(n0_over_2)
            edge_indices.extend([sorter[-1], sorter[-1]])

        else:
            nvec_sorted[-1] -= n1
            nvec_sorted[-2] = 0
            edge_reps.append(n1)
            edge_indices.extend([sorter[-1], sorter[-2]])

        sorterinv = _get_inverse_permutation(sorter)
        nvec = nvec_sorted[sorterinv]

    return np.array(edge_reps, dtype=dtype), np.array(edge_indices, dtype=dtype)


@nb.njit(cache=True)
def ix_(matrix, rows, cols):
    sliced_matrix = np.empty(shape=(len(rows), len(cols)), dtype=matrix.dtype)

    for idx in range(len(rows)):
        for jdx in range(len(cols)):
            sliced_matrix[idx, jdx] = matrix[rows[idx], cols[jdx]]

    return sliced_matrix


@nb.njit(cache=True)
def get_kept_edges(edge_reps, index):
    ret = np.empty_like(edge_reps)
    i = 0
    edge_reps_p_1 = edge_reps + 1

    for n in edge_reps_p_1:
        ret[i] = index % n
        index //= n
        i += 1

    return ret



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This file contains some common calculations for `BuiltinConnector`.
"""

import numpy as np
import numba as nb

from functools import lru_cache

from piquasso._math.indices import get_auxiliary_modes

from piquasso.fermionic._utils import (
    get_fock_subspace_dimension,
    cutoff_fock_space_dim_array,
    get_fock_subspace_index_first_quantized,
    next_first_quantized,
    get_fock_space_basis,
    get_fock_space_index,
)


def calculate_interferometer_on_fermionic_fock_space(connector, matrix, cutoff):
    """Calculates the representation of the interferometer matrix on the Fock space.

    This algorithm calculates the determinants of the submatrices recursively using
    Laplace's expansion. For this, it is easiest to work in the first quantized picture.
    """

    np = connector.forward_pass_np
    fallback_np = connector.fallback_np

    d = len(matrix)

    subspace_representations = []

    subspace_representations.append(np.array([[1.0]], dtype=matrix.dtype))

    if cutoff == 1:
        return subspace_representations

    subspace_representations.append(matrix)

    if cutoff == 2:
        return subspace_representations

    for n in range(2, cutoff):
        dim = get_fock_subspace_dimension(d, n)

        representation = np.zeros(shape=(dim, dim), dtype=matrix.dtype)

        previous_representation = subspace_representations[n - 1]

        first_quantized_row = fallback_np.arange(n)
        for row_idx in range(dim):
            first_quantized_col = fallback_np.arange(n)

            matrix_row_idx = first_quantized_row[0]

            deleted_row = first_quantized_row[1:]
            deleted_row_idx = get_fock_subspace_index_first_quantized(deleted_row, d)
            for col_idx in range(dim):
                sum_ = 0.0
                for laplace_index in range(n):
                    deleted_col = fallback_np.delete(first_quantized_col, laplace_index)
                    deleted_col_idx = get_fock_subspace_index_first_quantized(
                        deleted_col, d
                    )
                    sum_ += (
                        (-1) ** (laplace_index % 2)
                        * matrix[matrix_row_idx, first_quantized_col[laplace_index]]
                        * previous_representation[deleted_row_idx, deleted_col_idx]
                    )

                representation = connector.assign(
                    representation, (row_idx, col_idx), sum_
                )

                first_quantized_col = next_first_quantized(first_quantized_col, d)
            first_quantized_row = next_first_quantized(first_quantized_row, d)

        subspace_representations.append(representation)

    return subspace_representations


@nb.njit(cache=True)
def _nb_calculate_index_list_for_appling_interferometer(modes, d, cutoff):
    subspace = get_fock_space_basis(d=len(modes), cutoff=cutoff)
    auxiliary_subspace = get_fock_space_basis(d=d - len(modes), cutoff=cutoff)

    indices = cutoff_fock_space_dim_array(cutoff=np.arange(cutoff + 1), d=len(modes))
    auxiliary_indices = cutoff_fock_space_dim_array(
        cutoff=np.arange(cutoff + 1), d=d - len(modes)
    )
    auxiliary_modes = get_auxiliary_modes(d, modes)

    all_occupation_numbers = np.zeros(d, dtype=np.int64)

    index_list = []

    for n in range(cutoff):
        size = indices[n + 1] - indices[n]
        n_particle_subspace = subspace[indices[n] : indices[n + 1]]
        auxiliary_size = auxiliary_indices[cutoff - n]
        state_index_matrix = np.empty(shape=(size, auxiliary_size), dtype=np.int64)
        for idx1, auxiliary_occupation_numbers in enumerate(
            auxiliary_subspace[:auxiliary_size]
        ):
            for idx, mode in enumerate(auxiliary_modes):
                all_occupation_numbers[mode] = auxiliary_occupation_numbers[idx]

            for idx2, column_vector_on_subspace in enumerate(n_particle_subspace):
                for idx, mode in enumerate(modes):
                    all_occupation_numbers[mode] = column_vector_on_subspace[idx]

                column_index = get_fock_space_index(all_occupation_numbers)
                state_index_matrix[idx2, idx1] = column_index

        index_list.append(state_index_matrix)

    return index_list


_calculate_index_list_for_appling_interferometer = lru_cache(maxsize=None)(
    _nb_calculate_index_list_for_appling_interferometer
)


def apply_fermionic_passive_linear_to_state_vector(
    connector, representations, state_vector, modes, d, cutoff
):
    """Applies a passive linear gate to a state vector expressed in the Fock basis.

    This function assumes that the n-particle representations of the passive linear gate
    has already been calculated.
    """
    index_list = _calculate_index_list_for_appling_interferometer(modes, d, cutoff)[
        : len(representations)
    ]

    np = connector.forward_pass_np

    new_state_vector = np.zeros_like(state_vector)

    for n, indices in enumerate(index_list):
        new_state_vector = connector.assign(
            new_state_vector, indices, representations[n] @ state_vector[indices]
        )

    return new_state_vector



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Gates
=====

The built-in quantum gates of Piquasso.

Gates can be characterized by a unitary operator :math:`U`, which describes the
time evolution of quantum states as

.. math::
    \rho' = U \rho U^\dagger,

where :math:`\rho` is a density operator.

Gates with at most quadratic Hamiltonians are called linear gates. Evolution of the
ladder operators by linear gates could be expressed in the form

.. math::
    U^\dagger \boldsymbol{\xi} U = S_{(c)} \boldsymbol{\xi} + \boldsymbol{\beta},
    :label: linearity

where :math:`S_{(c)} \in \operatorname{Sp}(2d, \mathbb{R})` is a symplectic matrix in
the complex form,
:math:`\boldsymbol{\beta} = [\boldsymbol{\alpha}, \overline{\boldsymbol{\alpha}}]`,
:math:`\alpha \in \mathbb{C}^{d}`, and

.. math::
    \boldsymbol{\xi} := \begin{bmatrix}
        a_1, \dots, a_d, a_1^\dagger, \dots, a_d^\dagger
    \end{bmatrix}^T,
    :label: xi

where :math:`a_1^\dagger, ..., a_d^\dagger` and :math:`a_1, \dots, a_d` are the
creation and annihilation operators, respectively.

Most of the gates defined here are linear gates, which can be characterized by
:math:`S_{(c)}` and :math:`\boldsymbol{\beta}`.
"""

import abc

import numpy as np

from piquasso.api.instruction import Gate
from piquasso.api.exceptions import InvalidParameter

from piquasso._math.linalg import is_square, is_symmetric
from piquasso._math.symplectic import complex_symplectic_form, is_symplectic

from typing import Optional


class _PassiveLinearGate(Gate, abc.ABC):
    @abc.abstractmethod
    def _get_passive_block(self, connector, config):
        """
        The upper left (passive) block of the transformation.
        """


class _ActiveLinearGate(Gate, abc.ABC):
    @abc.abstractmethod
    def _get_passive_block(self, connector, config):
        """
        The upper left (passive) block of the transformation.
        """

    @abc.abstractmethod
    def _get_active_block(self, connector, config):
        """
        The upper right (active) block of the transformation.
        """


class Interferometer(_PassiveLinearGate):
    r"""Applies a general interferometer gate.

    The unitary operator corresponding to an interferometer can be written as

    .. math::
        I(U) = \exp(i \hat{H}),

    where :math:`\hat{H}` is a quadratic Hamiltonian

    .. math::
        \hat{H} = \boldsymbol{\xi}^\dagger H \boldsymbol{\xi}, \qquad H = \frac{1}{2}
        \begin{bmatrix}
            A & \\
            & \overline{A}
        \end{bmatrix},

    where :math:`A` is self-adjoint and :math:`\boldsymbol{\xi}` is defined in Eq.
    :eq:`xi`. The interferometer is parametrized as :math:`U = e^{iA}`.

    The interpretation of the interferometer matrix :math:`U` is that it is the
    one-particle unitary matrix, i.e., the matrix which acts on single particles.
    Moreover, the evolution of the ladder operators can be described by

    .. math::
        S_{(c)} = \begin{bmatrix}
            U & 0_{d \times d} \\
            0_{d \times d} & \overline{U}
        \end{bmatrix},

    where :math:`U \in \mathbb{C}^{d \times d}` is the interferometer matrix, and
    :math:`S_{(c)}` is defined by Eq. :eq:`linearity`.

    Args:
        matrix (numpy.ndarray):
            The unitary interferometer matrix, corresponding to a
            passive transformation on the one-particle subspace.
    """

    def __init__(self, matrix: np.ndarray) -> None:
        if not is_square(matrix):
            raise InvalidParameter(
                "The interferometer matrix should be a square matrix."
            )

        super().__init__(params=dict(matrix=matrix))

    def _get_passive_block(self, connector, config):
        return self._params["matrix"]


class Beamsplitter(_PassiveLinearGate):
    r"""Applies a beamsplitter gate.

    The unitary operator of the beamsplitter gate can be written as

    .. math::
        B_{ij} (\theta, \phi) = \exp \left (
            \theta e^{i \phi} a^\dagger_i a_j
            - \theta e^{- i \phi} a^\dagger_j a_i
        \right ).

    The beamsplitter transfer matrix is given by

    .. math::
        U = \begin{bmatrix}
            t & -\overline{r} \\
            r & t
        \end{bmatrix},

    where :math:`t = \cos(\theta)` and :math:`r = e^{i \phi} \sin(\theta)`.
    Moreover, the symplectic representation of the beamsplitter gate is

    .. math::
        S_{(c)} = U \oplus \overline{U} = \begin{bmatrix}
            t  & -\overline{r} &    & \\
            r & t   &    & \\
               &     & t  & -r \\
               &     & \overline{r} & t
        \end{bmatrix},

    where :math:`S_{(c)}` is defined by Eq. :eq:`linearity`.
    """

    NUMBER_OF_MODES = 2

    def __init__(self, theta: float = np.pi / 4, phi: float = 0.0) -> None:
        r"""
        Args:
            theta (float):
                The transmissivity angle of the beamsplitter.
                Defaults to :math:`\theta=\pi/4`, which yields a 50:50 beamsplitter.
            phi (float):
                Phase angle of the beamsplitter. Defaults to :math:`\phi = 0`.
        """

        super().__init__(
            params=dict(
                theta=theta,
                phi=phi,
            ),
        )

    def _get_passive_block(self, connector, config):
        np = connector.np

        theta = self._params["theta"]
        phi = self._params["phi"]

        t = np.cos(theta)
        r = np.exp(1j * phi) * np.sin(theta)

        return np.array(
            [
                [t, -np.conj(r)],
                [r, t],
            ],
            dtype=config.complex_dtype,
        )


class Beamsplitter5050(_PassiveLinearGate):
    r"""Applies a 50:50 beamsplitter gate.

    This gate corresponds to a regular beamsplitter gate :class:`Beamsplitter` with
    parameters :math:`\theta = \frac{\pi}{4}` and :math:`\phi = 0`.

    The beamsplitter transfer matrix is given by

    .. math::
        U = \frac{1}{\sqrt{2}} \begin{bmatrix}
            1 & -1 \\
            1 & 1
        \end{bmatrix}.
    """

    NUMBER_OF_MODES = 2

    def __init__(self) -> None:
        super().__init__()

    def _get_passive_block(self, connector, config):
        np = connector.np

        return np.array(
            [
                [1, -1],
                [1, 1],
            ],
            dtype=config.dtype,
        ) / np.sqrt(2)


class Phaseshifter(_PassiveLinearGate):
    r"""Applies a rotation or a phaseshifter gate.

    The unitary operator corresponding to the phaseshifter gate on the :math:`i`-th mode
    is

    .. math::
        R_i (\phi) = \exp \left (
            i \phi a_i^\dagger a_i
        \right ).

    The symplectic representation of the phaseshifter gate is

    .. math::
        S_{(c)} = \begin{bmatrix}
            e^{i \phi} & 0 \\
            0 & e^{- i \phi}
        \end{bmatrix},

    where :math:`S_{(c)}` is defined by Eq. :eq:`linearity`.
    """

    NUMBER_OF_MODES = 1

    def __init__(self, phi: float) -> None:
        """
        Args:
            phi (float): The angle of the rotation.
        """

        super().__init__(
            params=dict(phi=phi),
        )

    def _get_passive_block(self, connector, config):
        np = connector.np

        phi = self._params["phi"]

        return np.array([[np.exp(1j * phi)]], dtype=complex)


class MachZehnder(_PassiveLinearGate):
    r"""Mach-Zehnder interferometer.

    The Mach-Zehnder interferometer is equivalent to

    .. math::
        MZ(\phi_{int}, \phi_{ext}) =
            B(\pi/4, \pi/2) (R(\phi_{int}) \oplus \mathbb{1})
            B(\pi/4, \pi/2) (R(\phi_{ext}) \oplus \mathbb{1}).

    The symplectic representation of the Mach-Zehnder interferometer is

    .. math::
        S_{(c)} = \begin{bmatrix}
        e^{i \phi_{ext} } (e^{i \phi_{int} } - 1)   & i (e^{i \phi_{int} } + 1) & & \\
        i e^{i \phi_{ext} } (e^{i \phi_{int} } + 1) & 1 - e^{i \phi_{int} }     & & \\
        & & e^{-i \phi_{ext} } (e^{-i \phi_{int} } - 1) & -i (e^{-i \phi_{int} } + 1) \\
        & & -i e^{-i \phi_{ext} } (e^{-i \phi_{int} } + 1) & 1 - e^{-i \phi_{int} }
        \end{bmatrix},

    where :math:`\phi_{int}, \phi_{ext} \in \mathbb{R}` and :math:`S_{(c)}` is defined
    by Eq. :eq:`linearity`.
    """

    NUMBER_OF_MODES = 2

    def __init__(self, int_: float, ext: float) -> None:
        """
        Args:
            int_ (float): The internal angle.
            ext (float): The external angle.
        """

        super().__init__(
            params=dict(int_=int_, ext=ext),
        )

    def _get_passive_block(self, connector, config):
        np = connector.np

        int_ = self._params["int_"]
        ext = self._params["ext"]
        int_phase, ext_phase = np.exp(1j * np.array([int_, ext]))

        return (
            1
            / 2
            * connector.np.array(
                [
                    [ext_phase * (int_phase - 1), 1j * (int_phase + 1)],
                    [1j * ext_phase * (int_phase + 1), 1 - int_phase],
                ],
                dtype=config.complex_dtype,
            )
        )


class Fourier(_PassiveLinearGate):
    r"""Applies a Fourier gate.

    The Fourier gate simply transforms the quadratures as follows:

    .. math::
        F^\dagger \hat{x} F &= -\hat{p} \\
        F^\dagger \hat{p} F &= \hat{x}

    The unitary operator corresponding to the Fourier gate on the :math:`i`-th mode is

    .. math::
        F_{i} = \exp \left (
            i \frac{\pi}{2} a_i^\dagger a_i
        \right ).

    The symplectic representation of the Fourier gate is

    .. math::
        S_{(c)} = \begin{bmatrix}
            i & 0  \\
            0 & -i \\
        \end{bmatrix},

    where :math:`S_{(c)}` is defined by Eq. :eq:`linearity`.


    Note:
        Corresponds to the :class:`Phaseshifter` gate with :math:`\phi = \pi/2`.
    """

    NUMBER_OF_MODES = 1

    def __init__(self) -> None:
        super().__init__()

    def _get_passive_block(self, connector, config):
        return connector.np.array([[1j]], dtype=config.complex_dtype)


class GaussianTransform(_ActiveLinearGate):
    r"""Applies a Gaussian transformation gate.

    The symplectic representation of the Gaussian gate is

    .. math::
        S_{(c)} = \begin{bmatrix}
            P & A \\
            \overline{A} & \overline{P}
        \end{bmatrix},

    where :math:`P, A \in C^{d \times d}` fulfilling the relations

    .. math::
        P P^\dagger - A A^\dagger &= I_{d \times d}, \\
        P A^T &= A P^T,

    and :math:`S_{(c)}` is defined by Eq. :eq:`linearity`.
    """

    def __init__(self, passive: np.ndarray, active: np.ndarray) -> None:
        """
        Args:
            passive (numpy.ndarray):
                The passive submatrix :math:`P` of the symplectic matrix corresponding
                to the transformation.
            active (numpy.ndarray):
                The active submatrix :math:`A`
                of the symplectic matrix corresponding to the transformation.

        Raises:
            InvalidParameters: Raised if the parameters do not form a symplectic matrix.
        """
        if not is_symplectic(
            np.block([[passive, active], [active.conj(), passive.conj()]]),
            form_func=complex_symplectic_form,
        ):
            raise InvalidParameter(
                "The input parameters for instruction 'GaussianTransform' do not form "
                "a symplectic matrix."
            )

        super().__init__(params=dict(passive=passive, active=active))

    def _get_passive_block(self, connector, config):
        return self._params["passive"]

    def _get_active_block(self, connector, config):
        return self._params["active"]


class Squeezing(_ActiveLinearGate):
    r"""Applies the single-mode squeezing gate.

    The unitary operator corresponding to the squeezing gate is

    .. math::
        S_{i} (z) = \exp \left (
            \frac{1}{2}(\overline{z} a_i^2 - z a_i^{\dagger 2} )
        \right ).

    The symplectic representation of the squeezing gate is

    .. math::
        S_{(c)} = \begin{bmatrix}
            \cosh r & - e^{i \phi} \sinh r \\
            - e^{- i \phi} \sinh r & \cosh r
        \end{bmatrix},

    where :math:`S_{(c)}` is defined by Eq. :eq:`linearity`. The unitary operator
    corresponding to the squeezing gate is

    .. math::
        S(z) = \exp \left (
            \frac{1}{2}(\overline{z} a_i^2 - z a_i^{\dagger 2})
        \right ),

    where :math:`z = re^{i\phi}`.
    """

    NUMBER_OF_MODES = 1

    def __init__(self, r: float, phi: float = 0.0) -> None:
        """
        Args:
            r (float): The amplitude of the squeezing instruction.
            phi (float): The squeezing angle.
        """
        super().__init__(params=dict(r=r, phi=phi))

    def _get_passive_block(self, connector, config):
        np = connector.np

        r = self.params["r"]

        return np.array([[np.cosh(r)]], dtype=config.complex_dtype)

    def _get_active_block(self, connector, config):
        np = connector.np

        r = self.params["r"]
        phi = self.params["phi"]

        return np.array([[-np.sinh(r) * np.exp(1j * phi)]], dtype=config.complex_dtype)


class QuadraticPhase(_ActiveLinearGate):
    r"""Applies the quadratic phase instruction to the state.

    The unitary operator corresponding to the Fourier gate is

    .. math::
        QP_{i} (s) = \exp \left (
            i \frac{s}{2 \hbar} x_i^2
        \right ).

    The symplectic representation of the quadratic phase gate is

    .. math::
        S_{(c)} = \begin{bmatrix}
            1 + i \frac{s}{2} & i \frac{s}{2} \\
            -i \frac{s}{2} & 1 - i \frac{s}{2}
        \end{bmatrix},

    where :math:`S_{(c)}` is defined by Eq. :eq:`linearity`.
    """

    NUMBER_OF_MODES = 1

    def __init__(self, s: float) -> None:
        """
        Args:
            s (float): The parameter quadratic phase gate.
        """
        super().__init__(params=dict(s=s))

    def _get_passive_block(self, connector, config):
        s = self._params["s"]

        return connector.np.array([[1 + s / 2 * 1j]], dtype=config.complex_dtype)

    def _get_active_block(self, connector, config):
        s = self._params["s"]

        return connector.np.array([[s / 2 * 1j]], dtype=config.complex_dtype)


class Squeezing2(_ActiveLinearGate):
    r"""Applies the two-mode squeezing gate to the state.

    The unitary operator corresponding to the two-mode squeezing gate is

    .. math::
        S_{ij} (z) = \exp \left (
            \frac{1}{2}(\overline{z} a_i a_j - z a_i^\dagger a_j^\dagger )
        \right ).

    In the bosonic setting, symplectic representation of the two-mode squeezing gate is

    .. math::
        S_{(c)} = \begin{bmatrix}
            \cosh r & 0 & 0 & e^{i \phi} \sinh r \\
            0 & \cosh r & e^{i \phi} \sinh r & 0 \\
            0 & e^{- i \phi} \sinh r & \cosh r & 0 \\
            e^{- i \phi} \sinh r & 0 & 0 & \cosh r
        \end{bmatrix},

    where :math:`S_{(c)}` is defined by Eq. :eq:`linearity`.

    In the fermionic setting, an analoguous definition is used: the bosonic ladder
    operators are just exchanged to fermionic ones. For two fermionic modes, the action
    on :math:`\ket{00}` and :math:`\ket{11}` is simply given as

    .. math::
        S_{12} (z) \ket{00} &= \cos (r/2) \ket{00} - e^{i \phi} \sin (r/2) \ket{11}, \\
        S_{12} (z) \ket{11} &= \cos (r/2) \ket{11} + e^{-i \phi} \sin (r/2) \ket{00},

    where :math:`z = r e^{i \phi}`.
    """

    NUMBER_OF_MODES = 2

    def __init__(self, r: float, phi: float = 0.0) -> None:
        """
        Args:
            r (float): The amplitude of the squeezing instruction.
            phi (float): The squeezing angle.
        """
        super().__init__(params=dict(r=r, phi=phi))

    def _get_passive_block(self, connector, config):
        np = connector.np

        r = self._params["r"]

        return np.array(
            [
                [np.cosh(r), 0],
                [0, np.cosh(r)],
            ],
            dtype=config.complex_dtype,
        )

    def _get_active_block(self, connector, config):
        np = connector.np

        r = self._params["r"]
        phi = self._params["phi"]

        return np.array(
            [
                [0, np.sinh(r) * np.exp(1j * phi)],
                [np.sinh(r) * np.exp(1j * phi), 0],
            ],
            dtype=config.complex_dtype,
        )


class ControlledX(_ActiveLinearGate):
    r"""Applies the controlled-X gate to the state.

    The unitary operator corresponding to the controlled-X gate is

    .. math::
        CX_{ij} (s) = \exp \left (
            -i \frac{s}{\hbar} x_i p_j
        \right ).

    The symplectic representation of the controlled-X gate is

    .. math::
        S_{(c)} = \begin{bmatrix}
            1            & - \frac{s}{2} & 0           & \frac{s}{2} \\
            \frac{s}{2} & 1              & \frac{s}{2} & 0            \\
            0            & \frac{s}{2} & 1            & - \frac{s}{2} \\
            \frac{s}{2} & 0            & \frac{s}{2} & 1
        \end{bmatrix},

    where :math:`S_{(c)}` is defined by Eq. :eq:`linearity`.
    """

    NUMBER_OF_MODES = 2

    def __init__(self, s: float):
        """
        s (float): The parameter of the controlled-X gate.
        """
        super().__init__(params=dict(s=s))

    def _get_passive_block(self, connector, config):
        s = self._params["s"]

        return connector.np.array(
            [
                [1, -s / 2],
                [s / 2, 1],
            ],
            dtype=config.complex_dtype,
        )

    def _get_active_block(self, connector, config):
        s = self._params["s"]

        return connector.np.array(
            [
                [0, s / 2],
                [s / 2, 0],
            ],
            dtype=config.complex_dtype,
        )


class ControlledZ(_ActiveLinearGate):
    r"""Applies the controlled-Z gate to the state.

    The unitary operator corresponding to the controlled-Z gate is

    .. math::
        CZ_{ij} (s) = \exp \left (
            -i \frac{s}{\hbar} x_i x_j
        \right ).

    The symplectic representation of the controlled Z gate is

    .. math::
        S_{(c)} = \begin{bmatrix}
            1            & i \frac{s}{2} & 0            & i \frac{s}{2}    \\
            i \frac{s}{2} & 1              & i \frac{s}{2} & 0             \\
            0            & -i \frac{s}{2} & 1            & - i \frac{s}{2} \\
            -i \frac{s}{2} & 0            & -i \frac{s}{2} & 1
        \end{bmatrix},

    where :math:`S_{(c)}` is defined by Eq. :eq:`linearity`.
    """

    NUMBER_OF_MODES = 2

    def __init__(self, s: float):
        """
        s (float): The parameter of the controlled-Z gate.
        """
        super().__init__(params=dict(s=s))

    def _get_passive_block(self, connector, config):
        s = self._params["s"]

        return connector.np.array(
            [
                [1, 1j * (s / 2)],
                [1j * (s / 2), 1],
            ],
            dtype=config.complex_dtype,
        )

    def _get_active_block(self, connector, config):
        s = self._params["s"]

        return connector.np.array(
            [
                [0, 1j * (s / 2)],
                [1j * (s / 2), 0],
            ],
            dtype=config.complex_dtype,
        )


class Displacement(Gate):
    r"""Displacement gate.

    The displacement gate evolves the ladder operators by

    .. math::
        \boldsymbol{\xi} \mapsto \boldsymbol{\xi} + \begin{bmatrix}
            \alpha \\
            \overline{\alpha}
        \end{bmatrix},

    where :math:`\alpha \in \mathbb{C}` and :math:`\boldsymbol{\xi} = [a, a^\dagger]^T`
    is a formal vector of the ladder operators.

    One must either specify :math:`\alpha` only, or the combination of :math:`r` and
    :math:`\phi`.
    When :math:`r` and :math:`\phi` are the given parameters, :math:`\alpha` is
    calculated via:

    .. math::
        \alpha = r e^{i\phi}.

    For more intuition, one can consider transforming the quadrature operators
    :math:`\hat{x}_j` and :math:`\hat{p}_j`:

    .. math::
        D^\dagger_j(\alpha)
        \begin{bmatrix}
            \hat{x}_j \\
            \hat{p}_j
        \end{bmatrix}
        D_j(\alpha)
        = \begin{bmatrix}
            \hat{x}_j + \sqrt{2 \hbar} \mathrm{Re} \alpha I \\
            \hat{p}_j + \sqrt{2 \hbar} \mathrm{Im} \alpha I
        \end{bmatrix}.

    This means, that the parameter :math:`\alpha \in \mathbb{C}` describes the shift of
    the quadrature operators in the two-dimensional phase space, i.e., the real part is
    responsible for a shift in position, and the imaginary part for a shift in momentum.

    Note:
        The specified displacement is automatically scaled by :math:`\sqrt{2 \hbar}`.
    """

    NUMBER_OF_MODES = 1

    def __init__(
        self,
        r: float,
        phi: Optional[float] = 0.0,
    ) -> None:
        """
        Args:
            r (float): The magnitude of phase space displacement.
            phi (float): The angle of phase space displacement
        """

        super().__init__(params=dict(r=r, phi=phi))


class PositionDisplacement(Gate):
    r"""Position displacement gate. It affects only the :math:`\hat{x}` quadrature.

    Note:
        The specified displacement is automatically scaled by :math:`\sqrt{2 \hbar}`.
    """

    NUMBER_OF_MODES = 1

    def __init__(self, x: float) -> None:
        """
        Args:
            x (float): The position displacement.
        """
        super().__init__(params=dict(x=x), extra_params=dict(r=x, phi=0.0))


class MomentumDisplacement(Gate):
    r"""Momentum displacement gate. It only affects the :math:`\hat{p}` quadrature.

    Note:
        The specified displacement is automatically scaled by :math:`\sqrt{2 \hbar}`.
    """

    NUMBER_OF_MODES = 1

    def __init__(self, p: float) -> None:
        """
        Args:
            p (float): The momentum displacement.
        """
        super().__init__(params=dict(p=p), extra_params=dict(r=p, phi=np.pi / 2))


class CubicPhase(Gate):
    r"""Cubic Phase gate.

    The definition of the Cubic Phase gate is

    .. math::
        \operatorname{CP}(\gamma) = e^{i \hat{x}^3 \frac{\gamma}{3 \hbar}}

    The Cubic Phase gate transforms the annihilation operator as

    .. math::
        \operatorname{CP}^\dagger(\gamma) \hat{a} \operatorname{CP}(\gamma) = \hat{a}
            + i\frac{\gamma(\hat{a} +\hat{a}^\dagger)^2}{2\sqrt{2/\hbar}}

    It transforms the :math:`\hat{p}` quadrature as follows:

    .. math::
        \operatorname{CP}^\dagger(\gamma) \hat{p} \operatorname{CP}(\gamma) =
            \hat{p} + \gamma \hat{x}^2.

    Note:
        This is a non-linear gate, therefore it couldn't be used with
        :class:`~piquasso._simulators.gaussian.state.GaussianState`.
        Using this gate requires a high cutoff to make the more accurate simulation.
    """

    NUMBER_OF_MODES = 1

    def __init__(self, gamma: float) -> None:
        """
        Args:
            gamma (float): The Cubic Phase parameter.
        """
        super().__init__(params=dict(gamma=gamma))


class Kerr(Gate):
    r"""Kerr gate.

    The definition of the Kerr gate is

    .. math::
        K_i (\xi) = \exp \left (
            i \xi n_i n_i
        \right ).

    Since the Kerr gate is non-linear, the ladder operators are not evolved into a
    linear combination of ladder operators, which can be attributed to the fact that the
    commutator :math:`[a_j, \hat{n}^2_j]` is not a linear combination of the ladder
    operators, i.e., :math:`[a_j, \hat{n}^2_j] = a_j (2\hat{n}_j-I)`. Hence,
    the Kerr gate transforms the ladder operators as

    .. math::
        K^\dagger_j(\kappa) a_j K_j(\kappa) &=
            a_j \exp\left(i \kappa \left( 2 \hat{n}_j - I\right)\right),\\
        K^\dagger_j(\kappa) a_j^\dagger K_j(\kappa) &=
            a_j^\dagger \exp\left(
                -i \kappa \left(2 \hat{n}_j - I \right)
            \right).

    Note:
        This is a non-linear gate, therefore it cannot be used with
        :class:`~piquasso._simulators.gaussian.state.GaussianState`.
    """

    NUMBER_OF_MODES = 1

    def __init__(self, xi: float) -> None:
        """
        Args:
            xi (float): The magnitude of the Kerr nonlinear term.
        """
        super().__init__(params=dict(xi=xi))


class CrossKerr(Gate):
    r"""Cross-Kerr gate.

    The definition of the Cross-Kerr gate is

    .. math::
        CK_{ij} (\xi) = \exp \left (
            i \xi n_i n_j
        \right ).

    The Cross-Kerr gate transforms the annihilation operators as

    .. math::
        CK_{ij}^\dagger (\xi) a_i CK_{ij} (\xi) &= a_i \exp(- i \xi n_j) \\
        CK_{ij}^\dagger (\xi) a_j CK_{ij} (\xi) &= a_j \exp(- i \xi n_i)


    Note:
        This is a non-linear gate, therefore it couldn't be used with
        :class:`~piquasso._simulators.gaussian.state.GaussianState`.
    """

    NUMBER_OF_MODES = 2

    def __init__(self, xi: float) -> None:
        """
        Args:
            xi (float): The magnitude of the Cross-Kerr nonlinear term.
        """
        super().__init__(params=dict(xi=xi))


class Graph(Gate):
    r"""Applies a graph given its adjacency matrix, see
    `this article <https://arxiv.org/pdf/1612.01199.pdf>`_ for more details.
    It decomposes the adjacency matrix of a graph into single mode squeezers and
    interferometers.
    """

    def __init__(
        self, adjacency_matrix: np.ndarray, mean_photon_number: float = 1.0
    ) -> None:
        r"""
        Args:
            adjacency_matrix (numpy.ndarray): A symmetric matrix with a size of :math:`N
                \times N` and it can be real or complex.
            mean_photon_number (float, optional): The mean photon number :math:`\bar{n}`
                for a mode. Defaults to :math:`1.0`.
        Raises:
            InvalidParameter:
                If the adjacency matrix is not symmetric.
        """
        self.adjacency_matrix = adjacency_matrix

        if not is_symmetric(adjacency_matrix):
            raise InvalidParameter("The adjacency matrix should be symmetric.")

        super().__init__(
            params=dict(
                adjacency_matrix=adjacency_matrix,
                mean_photon_number=mean_photon_number,
            ),
        )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import numpy as np

from piquasso._math.functions import (
    gaussian_wigner_function,
    gaussian_wigner_function_for_scalar,
)


@pytest.fixture
def d():
    return 1


@pytest.fixture
def mean():
    return np.array([1, 2])


@pytest.fixture
def cov():
    return np.array(
        [
            [1, 2],
            [-3, 4],
        ]
    )


def test_wigner_function_at_scalar(d, mean, cov):
    quadrature_array = np.array([1, 2])

    expected = 0.10065842420897406

    actual = gaussian_wigner_function_for_scalar(
        X=quadrature_array, d=d, mean=mean, cov=cov
    )

    assert np.allclose(expected, actual)


def test_gaussian_wigner_function_handles_vectors(d, mean, cov):
    positions = [[1.0], [3.0], [5.0]]
    momentums = [[2.0], [4.0], [6.0]]

    expected = np.array(
        [
            [0.10065842420897406, 0.020322585354620785, 0.00016724973685064803],
            [0.0674733595496344, 0.009131526225575573, 5.0374652683254064e-05],
            [0.020322585354620785, 0.001843623348920587, 6.81746788883418e-06],
        ]
    )

    actual = gaussian_wigner_function(positions, momentums, d=d, mean=mean, cov=cov)

    assert np.allclose(expected, actual)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This file contains some common calculations for `NumpyConnector`.
"""

import numpy as np

import numba as nb

from .._utils import precalculate_fermionic_passive_linear_indices


@nb.njit(cache=True)
def calculate_interferometer_on_fermionic_fock_space(matrix, cutoff):
    """Calculates the representation of the interferometer matrix on the Fock space.

    This algorithm calculates the determinants of the submatrices recursively using
    Laplace's expansion. For this, it is easiest to work in the first quantized picture.
    """

    d = len(matrix)

    subspace_representations = []

    subspace_representations.append(np.array([[1.0]], dtype=matrix.dtype))

    if cutoff == 1:
        return subspace_representations

    subspace_representations.append(matrix)

    if cutoff == 2:
        return subspace_representations

    for n in range(2, cutoff):
        laplace_indices, deleted_indices = (
            precalculate_fermionic_passive_linear_indices(n, d)
        )

        dim = laplace_indices.shape[0]

        representation = np.zeros(shape=(dim, dim), dtype=matrix.dtype)

        previous_representation = subspace_representations[n - 1]

        for row_idx in range(dim):
            matrix_row_idx = laplace_indices[row_idx, 0]
            deleted_row_idx = deleted_indices[row_idx, 0]

            for col_idx in range(dim):
                for laplace_index in range(n):
                    deleted_col_idx = deleted_indices[col_idx, laplace_index]

                    representation[row_idx, col_idx] += (
                        (-1) ** (laplace_index % 2)
                        * matrix[
                            matrix_row_idx, laplace_indices[col_idx, laplace_index]
                        ]
                        * previous_representation[deleted_row_idx, deleted_col_idx]
                    )

        subspace_representations.append(representation)

    return subspace_representations



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

import numba as nb

from piquasso.fermionic._utils import (
    get_fock_subspace_dimension,
    get_fock_subspace_index_first_quantized,
    next_first_quantized,
)


@nb.njit(cache=True)
def precalculate_fermionic_passive_linear_indices(n, d):
    dim = get_fock_subspace_dimension(d, n)

    laplace_indices = np.empty((dim, n), dtype=np.int64)
    deleted_indices = np.empty((dim, n), dtype=np.int64)

    first_quantized = np.arange(n)
    for row_idx in range(dim):
        for laplace_index in range(n):
            deleted = np.delete(first_quantized, laplace_index)
            deleted_idx = get_fock_subspace_index_first_quantized(deleted, d)
            deleted_indices[row_idx, laplace_index] = deleted_idx
            laplace_indices[row_idx, laplace_index] = first_quantized[laplace_index]

        first_quantized = next_first_quantized(first_quantized, d)

    return (laplace_indices, deleted_indices)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import piquasso as pq


@pytest.mark.parametrize("SimulatorClass", (pq.FockSimulator, pq.PureFockSimulator))
def test_FockState_get_particle_detection_probability(SimulatorClass):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=0.1, phi=np.pi / 3)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 4)

    simulator = SimulatorClass(d=2, config=pq.Config(cutoff=4))
    state = simulator.execute(program).state
    state.normalize()

    probability = state.get_particle_detection_probability(occupation_number=(0, 2))

    assert np.isclose(probability, 0.0012355767142126952)


@pytest.mark.parametrize("SimulatorClass", (pq.FockSimulator, pq.PureFockSimulator))
def test_FockState_quadratures_mean_variance(SimulatorClass):
    alpha = 0.2 - 0.2j

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=np.abs(alpha), phi=np.angle(alpha))
        pq.Q(0) | pq.Squeezing(r=0.2)

    simulator = SimulatorClass(d=1, config=pq.Config(cutoff=6))
    state = simulator.execute(program).state
    state.normalize()

    mean, var = state.quadratures_mean_variance(modes=(0,))

    assert np.isclose(mean, 0.3275267, atol=1e-5)
    assert np.isclose(var, 0.6709301, atol=1e-4)


@pytest.mark.parametrize("SimulatorClass", (pq.FockSimulator, pq.PureFockSimulator))
def test_FockState_wigner_function(SimulatorClass):
    alpha = 1 - 0.5j

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=np.abs(alpha), phi=np.angle(alpha))
        pq.Q(0) | pq.Squeezing(r=0.2)

    simulator = SimulatorClass(d=1, config=pq.Config(cutoff=10))
    state = simulator.execute(program).state
    state.normalize()

    wigner_function_values = state.wigner_function(
        positions=[1, 1.1],
        momentums=[-0.5, -0.6],
    )

    assert np.allclose(
        wigner_function_values,
        np.array([[0.09872782, 0.10777084], [0.10327266, 0.11273187]]),
    )


@pytest.mark.parametrize("SimulatorClass", (pq.FockSimulator, pq.PureFockSimulator))
def test_FockState_wigner_function_plot(SimulatorClass):
    config = pq.Config(cutoff=10, hbar=42)

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=0.10, phi=np.angle(1 - 0.5j))
        pq.Q(0) | pq.Squeezing(r=0.10)

    fock_simulator = SimulatorClass(d=1, config=config)
    fock_state = fock_simulator.execute(program).state
    x = np.linspace(-5, 5, 20)
    p = np.linspace(-5, 5, 20)
    mode = 0
    try:
        fock_state.plot_wigner(x, p, mode=mode)
    except Exception as e:
        pytest.fail(f"Plotting failed with exception: {e}")


@pytest.mark.parametrize("SimulatorClass", (pq.FockSimulator, pq.PureFockSimulator))
def test_FockState_plot_wigner_function_raises_InvalidModes_for_multiple_modes(
    SimulatorClass,
):
    alpha = 1 - 0.5j

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=np.abs(alpha), phi=np.angle(alpha))
        pq.Q(0) | pq.Squeezing(r=0.2)

    simulator = SimulatorClass(d=1)
    state = simulator.execute(program).state

    with pytest.raises(pq.api.exceptions.InvalidModes):
        x = np.linspace(-5, 5, 20)
        p = np.linspace(-5, 5, 20)
        modes = 2
        state.plot_wigner(x, p, mode=modes)


@pytest.mark.parametrize("SimulatorClass", (pq.FockSimulator, pq.PureFockSimulator))
def test_FockState_plot_wigner_function_raises_InvalidModes_for_multiple_dimensions(
    SimulatorClass,
):
    alpha = 1 - 0.5j

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=np.abs(alpha), phi=np.angle(alpha))
        pq.Q(0) | pq.Squeezing(r=0.2)

    simulator = SimulatorClass(d=2)
    state = simulator.execute(program).state

    with pytest.raises(pq.api.exceptions.InvalidModes):
        x = np.linspace(-5, 5, 20)
        p = np.linspace(-5, 5, 20)
        state.plot_wigner(x, p)


@pytest.mark.parametrize("SimulatorClass", (pq.FockSimulator, pq.PureFockSimulator))
def test_FockState_wigner_function_raises_InvalidModes_for_multiple_modes(
    SimulatorClass,
):
    alpha = 1 - 0.5j

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=np.abs(alpha), phi=np.angle(alpha))
        pq.Q(0) | pq.Squeezing(r=0.2)

    simulator = SimulatorClass(d=2)
    state = simulator.execute(program).state

    with pytest.raises(pq.api.exceptions.InvalidModes):
        state.wigner_function(
            positions=[1, 1.1],
            momentums=[-0.5, -0.6],
        )


@pytest.mark.parametrize("SimulatorClass", (pq.FockSimulator, pq.PureFockSimulator))
def test_FockState_wigner_function_raises_InvalidModes_for_multiple_modes_specified(
    SimulatorClass,
):
    alpha = 1 - 0.5j

    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Displacement(r=np.abs(alpha), phi=np.angle(alpha))
        pq.Q(0) | pq.Squeezing(r=0.2)

    simulator = SimulatorClass(d=2)
    state = simulator.execute(program).state

    with pytest.raises(pq.api.exceptions.InvalidModes):
        state.wigner_function(
            positions=[1, 1.1],
            momentums=[-0.5, -0.6],
        )


@pytest.mark.parametrize("SimulatorClass", (pq.FockSimulator, pq.PureFockSimulator))
def test_FockState_fidelity(
    SimulatorClass,
):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=0.2, phi=np.pi / 2)

    simulator = SimulatorClass(d=1, config=pq.Config(cutoff=6))
    state = simulator.execute(program).state
    state.normalize()

    with pq.Program() as program_2:
        pq.Q() | pq.Vacuum()

        pq.Q(0) | pq.Squeezing(r=0.2, phi=-np.pi / 2)

    state_2 = simulator.execute(program_2).state
    state_2.normalize()
    fid = state.fidelity(state_2)

    assert np.isclose(fid, 0.92507584)
    assert np.isclose(state_2.fidelity(state), fid)
    assert np.isclose(state_2.fidelity(state_2), 1.0)


def test_FockState_get_purity():
    with pq.Program() as program:
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 1), bra=(0, 0, 1)) / 4
        pq.Q() | pq.DensityMatrix(ket=(0, 0, 2), bra=(0, 0, 2)) / 4
        pq.Q() | pq.DensityMatrix(ket=(0, 1, 1), bra=(0, 1, 1)) / 2

        pq.Q() | pq.DensityMatrix(ket=(0, 0, 2), bra=(0, 1, 1)) * np.sqrt(1 / 8)
        pq.Q() | pq.DensityMatrix(ket=(0, 1, 1), bra=(0, 0, 2)) * np.sqrt(1 / 8)

    simulator = pq.FockSimulator(d=3, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    state.validate()

    purity = state.get_purity()

    assert np.isclose(purity, 0.625)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import numba as nb

from .powtrace import calc_power_traces

from piquasso._math.combinatorics import comb

from .utils import match_occupation_numbers, ix_, get_kept_edges


@nb.njit(cache=True)
def _scale_matrix(matrix):
    r"""
    Scales the matrix for better accuracy.
    """

    dim = matrix.shape[0]
    scale_factor = np.sum(np.abs(matrix) ** 2)

    if scale_factor < 1e-8:
        return matrix, 1.0

    scale_factor = np.sqrt(scale_factor / 2) / (dim**2)

    matrix *= scale_factor

    return matrix, scale_factor


@nb.njit(cache=True)
def _calc_reduced_matrix(matrix: np.ndarray, delta: np.ndarray) -> np.ndarray:
    r"""
    Creates the input data required to calculate power traces according to `delta`.

    Args:
        matrix (np.ndarray): The input matrix.
        delta (np.ndarray): Elements required in $X_{\delta}$ from Eq. (B11) from
            https://arxiv.org/abs/2108.01622.

    Returns:
        Tuple: The input data required to calculate power traces.
    """
    nonzero_indices = np.nonzero(delta)[0]
    nonzero_dim_over_2 = len(nonzero_indices)

    nonzero_dim = 2 * nonzero_dim_over_2

    reduced_matrix = np.empty((nonzero_dim, nonzero_dim), dtype=matrix.dtype)

    for i in range(nonzero_dim_over_2):
        even = 2 * i
        odd = 2 * i + 1

        delta_i = delta[nonzero_indices[i]]

        source_col_even = 2 * nonzero_indices[i] + 1
        source_col_odd = 2 * nonzero_indices[i]

        for row_idx in range(nonzero_dim):
            source_row = 2 * nonzero_indices[row_idx // 2] + row_idx % 2

            reduced_matrix[row_idx, even] = (
                delta_i * matrix[source_row, source_col_even]
            )
            reduced_matrix[row_idx, odd] = delta_i * matrix[source_row, source_col_odd]

    return reduced_matrix


@nb.njit(cache=True)
def _calc_f(traces, scale_factor):
    r"""Calcualates `f` from Appendix B.1 from https://arxiv.org/abs/2108.01622.

    This function uses the previously calculated power traces.

    Args:
        traces (numpy.ndarray): The power traces.
        scale_factor (float): The scale factor of the input matrix for the power traces.
    """

    dim_over_2 = len(traces)
    dim = 2 * dim_over_2

    aux0 = np.zeros(dim_over_2 + 1, dtype=traces.dtype)
    aux1 = np.zeros(dim_over_2 + 1, dtype=traces.dtype)

    aux0[0] = 1.0

    data = [aux0, aux1]

    p_aux0 = 0
    p_aux1 = 1

    inverse_scale_factor = 1 / scale_factor
    for idx in range(1, dim_over_2 + 1):
        factor = traces[idx - 1] * inverse_scale_factor / (2.0 * idx)

        powfactor = 1.0

        if idx % 2 == 1:
            p_aux0 = 0
            p_aux1 = 1
        else:
            p_aux0 = 1
            p_aux1 = 0

        data[p_aux1] = np.copy(data[p_aux0])

        for jdx in range(1, dim // (2 * idx) + 1):
            powfactor = powfactor * factor / jdx
            start_idx = idx * jdx

            for kdx in range(start_idx, dim_over_2 + 1):
                data[p_aux1][kdx] += data[p_aux0][kdx - start_idx] * powfactor

        inverse_scale_factor = inverse_scale_factor / scale_factor

    return data[p_aux1]


@nb.njit(cache=True, parallel=True)
def hafnian_with_reduction(matrix_orig, occupation_numbers):
    r"""
    Calculates the hafnian of the input matrix using the power trace algorithm with
    Glynn-type iterations.

    The algorithm is enhanced by factoring in repetitions according to
    https://arxiv.org/abs/2108.01622.
    """
    n = sum(occupation_numbers)

    if n == 0:
        return 1.0
    elif n % 2 != 0:
        return 0.0

    all_edges, edge_indices = match_occupation_numbers(occupation_numbers)

    matrix_reduced = ix_(matrix_orig, edge_indices, edge_indices)

    if matrix_reduced.shape[0] <= 10:
        scale_factor = 1.0
        matrix = matrix_reduced
    else:
        scale_factor = (
            np.sum(np.abs(matrix_reduced)) / matrix_reduced.shape[0] ** 2 / np.sqrt(2.0)
        )
        matrix = matrix_reduced / scale_factor

    dim_over_2 = np.sum(all_edges)
    number_of_reps = len(all_edges)

    result = 0.0

    size = np.prod(all_edges + 1) // 2

    comb_cache = {}

    for n in range(max(all_edges) + 1):
        for k in range(n + 1):
            comb_cache[(n, k)] = comb(n, k)

    for permutation_idx in nb.prange(size):
        kept_edges = get_kept_edges(all_edges, permutation_idx)
        fact = False
        combinatorial_factor = 1.0

        delta = np.empty_like(all_edges)

        for i in range(number_of_reps):
            no_of_edges = all_edges[i]
            no_of_kept_edges = kept_edges[i]

            fact ^= (no_of_edges - no_of_kept_edges) % 2

            comb_input = (no_of_edges, no_of_kept_edges)
            combinatorial_factor *= comb_cache[comb_input]

            delta[i] = 2 * no_of_kept_edges - no_of_edges

        prefactor = (-1 if fact else 1) * combinatorial_factor

        reduced_matrix = _calc_reduced_matrix(matrix, delta)

        reduced_matrix, second_scale_factor = _scale_matrix(reduced_matrix)

        traces = calc_power_traces(reduced_matrix, dim_over_2)

        summand = prefactor * _calc_f(traces, second_scale_factor)[dim_over_2]

        result += summand

    result *= np.power(scale_factor, dim_over_2)
    result /= 1 << (dim_over_2 - 1)

    return result


@nb.njit(cache=True, parallel=True)
def hafnian_with_reduction_batch(matrix_orig, occupation_numbers, cutoff):
    r"""
    Calculates the hafnian of the input matrix using the power trace algorithm with
    Glynn-type iterations with batching, as described in
    https://arxiv.org/abs/2108.01622.
    """
    particle_number_sum = sum(occupation_numbers)

    odd = particle_number_sum % 2

    if odd:
        result_size = cutoff // 2 - 1
        new_occupation_numbers = np.copy(occupation_numbers)
        new_occupation_numbers[-1] += 1
    else:
        result_size = (cutoff - 1) // 2
        new_occupation_numbers = occupation_numbers

    all_edges, edge_indices = match_occupation_numbers(new_occupation_numbers)

    new_all_edges = np.empty(len(all_edges) + 1, dtype=all_edges.dtype)
    new_all_edges[: len(all_edges)] = all_edges
    new_all_edges[-1] = result_size
    all_edges = new_all_edges

    new_edge_indices = np.empty(len(edge_indices) + 2, dtype=edge_indices.dtype)
    new_edge_indices[: len(edge_indices)] = edge_indices
    new_edge_indices[-1] = len(matrix_orig) - 1
    new_edge_indices[-2] = len(matrix_orig) - 1
    edge_indices = new_edge_indices

    matrix_reduced = ix_(matrix_orig, edge_indices, edge_indices)

    if matrix_reduced.shape[0] <= 10:
        scale_factor = 1.0
        matrix = matrix_reduced
    else:
        scale_factor = (
            np.sum(np.abs(matrix_reduced)) / matrix_reduced.shape[0] ** 2 / np.sqrt(2.0)
        )
        matrix = matrix_reduced / scale_factor

    dim_over_2 = np.sum(all_edges)
    number_of_reps = len(all_edges)

    size = np.prod(all_edges + 1) // 2

    comb_cache = {}

    for n in range(max(all_edges) + 1):
        for k in range(n + 1):
            comb_cache[(n, k)] = comb(n, k)

    result = np.zeros(shape=result_size + 1, dtype=matrix.dtype)

    for permutation_idx in nb.prange(size):
        summand = np.zeros(shape=result_size + 1, dtype=matrix.dtype)
        kept_edges = get_kept_edges(all_edges, permutation_idx)
        fact = False ^ (result_size % 2)
        combinatorial_factor = 1.0

        delta = np.empty_like(all_edges)

        for i in range(number_of_reps):
            no_of_edges = all_edges[i]
            no_of_kept_edges = kept_edges[i]

            fact ^= (no_of_edges - no_of_kept_edges) % 2

            if i != number_of_reps - 1:
                comb_input = (no_of_edges, no_of_kept_edges)
                combinatorial_factor *= comb_cache[comb_input]

            delta[i] = 2 * no_of_kept_edges - no_of_edges

        reduced_matrix = _calc_reduced_matrix(matrix, delta)

        reduced_matrix, second_scale_factor = _scale_matrix(reduced_matrix)

        traces = calc_power_traces(reduced_matrix, dim_over_2)

        fs = _calc_f(traces, second_scale_factor)

        for idx in range(kept_edges[-1], result_size + 1):
            prefactor = (
                (-1 if fact ^ (idx % 2) else 1)
                * combinatorial_factor
                * comb_cache[idx, kept_edges[-1]]
            )

            if idx >= result_size - kept_edges[-1]:
                prefactor *= (
                    1
                    + (-1 if (result_size - idx) % 2 else 1)
                    * comb_cache[idx, result_size - kept_edges[-1]]
                    / comb_cache[idx, kept_edges[-1]]
                )

            summand[idx] = prefactor * fs[dim_over_2 - result_size + idx]

        result += summand

    for i in range(result_size + 1):
        result[i] *= np.power(scale_factor, dim_over_2 + i - result_size)
        result[i] /= 1 << (dim_over_2 - result_size + i)

    concat_result = np.zeros(cutoff, dtype=result.dtype)

    for i in range(len(result)):
        concat_result[2 * i + odd] = result[i]

    if particle_number_sum == 0:
        concat_result[0] = 1.0

    return concat_result



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import piquasso as pq


@pytest.mark.parametrize(
    "SimulatorClass",
    (
        pq.GaussianSimulator,
        pq.PureFockSimulator,
        pq.FockSimulator,
        pq.SamplingSimulator,
    ),
)
def test_get_particle_detection_probability_raises_PiquassoException_wrong_modes(
    SimulatorClass,
):
    wrong_occupation_number = (0, 1)

    with pq.Program() as program:
        pq.Q(0) | pq.Phaseshifter(np.pi / 3)

    simulator = SimulatorClass(d=3)

    state = simulator.execute(program).state

    with pytest.raises(pq.api.exceptions.PiquassoException):
        state.get_particle_detection_probability(
            occupation_number=wrong_occupation_number
        )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import numba as nb


@nb.njit(parallel=True, cache=True)
def calculate_interferometer_on_fock_space(interferometer, helper_indices):
    cutoff = len(helper_indices[0]) + 2
    subspace_representations = []

    subspace_representations.append(np.array([[1.0]], dtype=interferometer.dtype))
    subspace_representations.append(interferometer)

    for n in range(2, cutoff):
        subspace_indices = helper_indices[0][n - 2]
        first_subspace_indices = helper_indices[2][n - 2]

        first_nonzero_indices = helper_indices[1][n - 2]

        sqrt_occupation_numbers = helper_indices[3][n - 2]
        sqrt_first_occupation_numbers = helper_indices[4][n - 2]

        previous_representation = subspace_representations[n - 1]

        result_shape = (
            first_nonzero_indices.shape[0],
            sqrt_occupation_numbers.shape[0],
        )

        representation = np.zeros(result_shape, dtype=interferometer.dtype)

        for k in range(result_shape[0]):
            denominator = sqrt_first_occupation_numbers[k]
            previous_representation_indexed = previous_representation[
                first_subspace_indices[k]
            ]

            for j in range(sqrt_occupation_numbers.shape[1]):
                one_particle_contrib = (
                    interferometer[first_nonzero_indices[k], j] / denominator
                )

                for i in range(result_shape[1]):
                    representation[k, i] += (
                        one_particle_contrib
                        * sqrt_occupation_numbers[i, j]
                        * previous_representation_indexed[subspace_indices[i, j]]
                    )

        subspace_representations.append(representation.astype(interferometer.dtype))

    return subspace_representations



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple

import numpy as np
from piquasso._simulators.sampling.state import SamplingState

from piquasso.instructions import gates

from piquasso.api.exceptions import InvalidState, NotImplementedCalculation
from piquasso.api.result import Result
from piquasso.api.instruction import Instruction

from piquasso._simulators.fock.pure.calculations import (
    post_select_photons as pure_fock_post_select_photons,
    imperfect_post_select_photons as pure_fock_imperfect_post_select_photons,
)


from .utils import (
    generate_lossless_samples,
    generate_uniform_lossy_samples,
    generate_lossy_samples,
)


def state_vector(state: SamplingState, instruction: Instruction, shots: int) -> Result:
    coefficient = instruction._all_params["coefficient"]

    if "occupation_numbers" in instruction._all_params:
        occupation_numbers = instruction._all_params["occupation_numbers"]

        if state._config.validate and len(occupation_numbers) != state.d:
            raise InvalidState(
                f"The occupation numbers '{occupation_numbers}' are not well-defined "
                f"on '{state.d}' modes: instruction={instruction}"
            )

        state._occupation_numbers.append(np.rint(occupation_numbers).astype(int))
        state._coefficients.append(coefficient)

    elif "fock_amplitude_map" in instruction._all_params:
        for occupation_numbers, amplitude in instruction._all_params[
            "fock_amplitude_map"
        ].items():
            if state._config.validate and len(occupation_numbers) != state.d:
                raise InvalidState(
                    f"The occupation numbers '{occupation_numbers}' "
                    f"are not well-defined "
                    f"on '{state.d}' modes: instruction={instruction}"
                )

            state._occupation_numbers.append(np.rint(occupation_numbers).astype(int))
            state._coefficients.append(coefficient * amplitude)

    return Result(state=state)


def passive_linear(
    state: SamplingState, instruction: gates._PassiveLinearGate, shots: int
) -> Result:
    r"""Applies an interferometer to the circuit.

    This can be interpreted as placing another interferometer in the network, just
    before performing the sampling. This instruction is realized by multiplying
    current effective interferometer matrix with new interferometer matrix.

    Do note, that new interferometer matrix works as interferometer matrix on
    qumodes (provided as the arguments) and as an identity on every other mode.
    """
    _apply_matrix_on_modes(
        state=state,
        matrix=instruction._get_passive_block(state._connector, state._config),
        modes=instruction.modes,
    )

    return Result(state=state)


def _apply_matrix_on_modes(
    state: SamplingState, matrix: np.ndarray, modes: Tuple[int, ...]
) -> None:
    connector = state._connector
    np = connector.np
    fallback_np = connector.fallback_np

    embedded = np.identity(len(state.interferometer), dtype=state._config.complex_dtype)

    embedded = connector.assign(embedded, fallback_np.ix_(modes, modes), matrix)

    state.interferometer = embedded @ state.interferometer


def loss(state: SamplingState, instruction: Instruction, shots: int) -> Result:
    state.is_lossy = True

    _apply_matrix_on_modes(
        state=state,
        matrix=np.array([[instruction._all_params["transmissivity"]]]),
        modes=instruction.modes,
    )

    return Result(state=state)


def lossy_interferometer(
    state: SamplingState, instruction: Instruction, shots: int
) -> Result:
    state.is_lossy = True

    _apply_matrix_on_modes(
        state=state,
        matrix=instruction._all_params["matrix"],
        modes=instruction.modes,
    )

    return Result(state=state)


def particle_number_measurement(
    state: SamplingState, instruction: Instruction, shots: int
) -> Result:
    """
    Simulates a boson sampling using generalized Clifford & Clifford algorithm
    from [Brod, Oszmaniec 2020] see
    `this article <https://arxiv.org/abs/1906.06696>`_ for more details.

    This is a contribution implementation from `theboss`, see
    https://github.com/Tomev-CTP/theboss.

    This method assumes that initial_state is given in the second quantization
    description (mode occupation).

    Generalized Cliffords simulation strategy form [Brod, Oszmaniec 2020] was used
    as it allows effective simulation of broader range of input states than original
    algorithm.
    """

    if (
        state._config.validate
        and len(state._occupation_numbers) != 1
        and not np.isclose(state._coefficients[0], 1.0)
    ):
        raise NotImplementedCalculation(
            f"The instruction {instruction} is not supported for states defined using "
            "multiple 'StateVector' instructions.\n"
            "If you need this feature to be implemented, please create an issue at "
            "https://github.com/Budapest-Quantum-Computing-Group/piquasso/issues"
        )

    initial_state = state._occupation_numbers[0]

    interferometer_svd = np.linalg.svd(state.interferometer)

    singular_values = interferometer_svd[1]

    if not state.is_lossy:
        samples = generate_lossless_samples(
            initial_state,
            shots,
            state._connector.permanent,
            state.interferometer,
            state._config.rng,
        )
    elif np.all(np.isclose(singular_values, singular_values[0])):
        uniform_transmission_probability = singular_values[0] ** 2

        samples = generate_uniform_lossy_samples(
            initial_state,
            shots,
            state._connector.permanent,
            state.interferometer,
            uniform_transmission_probability,
            state._config.rng,
        )

    else:
        samples = generate_lossy_samples(
            initial_state,
            shots,
            state._connector.permanent,
            interferometer_svd,
            state._config.rng,
        )

    return Result(state=state, samples=list(map(tuple, samples)))


post_select_photons = pure_fock_post_select_photons
imperfect_post_select_photons = pure_fock_imperfect_post_select_photons



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import time

from scipy.stats import unitary_group
import matplotlib.pyplot as plt

import strawberryfields as sf

import json

import numpy as np

import piquasso as pq


N = 100  # number of shots


def get_strawberry_samples(squeezings, unitary):
    d = len(squeezings)

    sf_program = sf.Program(d)
    sf_engine = sf.Engine(backend="gaussian")

    with sf_program.context as q:
        for i in range(d):
            sf.ops.Sgate(squeezings[i]) | q[i]

        sf.ops.Interferometer(unitary) | tuple(q[i] for i in range(d))

        sf.ops.MeasureFock() | tuple(q[i] for i in range(d))

    return sf_engine.run(sf_program, shots=N).samples


def get_piquasso_samples(squeezings, unitary):
    d = len(squeezings)

    with pq.Program() as program:
        for i in range(d):
            pq.Q(i) | pq.Squeezing(squeezings[i])

        pq.Q() | pq.Interferometer(unitary)
        pq.Q() | pq.ParticleNumberMeasurement()

    simulator = pq.GaussianSimulator(d=d, config=pq.Config(measurement_cutoff=5 + 1))

    result = simulator.execute(program, shots=N)

    return result.samples


if __name__ == "__main__":
    FILENAME = f"gaussian_boson_sampling_{int(time.time())}.json"

    np.random.seed(123)

    # Warmup
    m = 2
    squeezings = np.arcsinh(1)
    unitary = unitary_group.rvs(m)
    get_strawberry_samples(squeezings, unitary)
    get_piquasso_samples(squeezings, unitary)
    ####

    x = []
    sf_times = []
    pq_times = []

    for m in range(2, 100):
        print("m=", m)
        x.append(m)

        squeezings = np.arcsinh(1)
        unitary = unitary_group.rvs(m)

        start_time = time.time()
        samples = get_strawberry_samples(squeezings, unitary)
        runtime = time.time() - start_time
        print("SF:", runtime)
        sf_times.append(runtime)

        start_time = time.time()
        samples = get_piquasso_samples(squeezings, unitary)
        runtime = time.time() - start_time
        print("PQ:", runtime)
        pq_times.append(runtime)

        with open(FILENAME, "w") as f:
            json.dump(dict(x=x, pq_times=pq_times, sf_times=sf_times), f, indent=4)

    plt.scatter(x, sf_times, label="Strawberry Fields")
    plt.scatter(x, pq_times, label="Piquasso")

    plt.yscale("log")

    plt.legend()
    plt.show()



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, List, Type

from piquasso.api.config import Config
from piquasso.api.connector import BaseConnector

from piquasso.api.simulator import Simulator
from piquasso.api.exceptions import InvalidSimulation

from .connectors.connector import BuiltinConnector


class BuiltinSimulator(Simulator):
    _extra_builtin_connectors: List[Type[BaseConnector]] = []

    def __init__(
        self,
        d: int,
        config: Optional[Config] = None,
        connector: Optional[BaseConnector] = None,
    ) -> None:
        if connector is not None:
            self._validate_connector(connector)

        super().__init__(d=d, config=config, connector=connector)

    def _validate_connector(self, connector):
        if (
            isinstance(connector, BuiltinConnector)
            and not isinstance(connector, self._default_connector_class)
            and not any(
                [isinstance(connector, cls) for cls in self._extra_builtin_connectors]
            )
        ):
            raise InvalidSimulation(
                f"The connector '{connector}' is not supported."
                f"Supported connectors:"
                "\n"
                f"{[self._default_connector_class] + self._extra_builtin_connectors}"
            )
